/****************************************************************************
 *
 * Copyright (c) 2001-2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

#include <config.h>

#include <xpl.h>
#include <memmgr.h>
#include <logger.h>
#include <hulautil.h>
#include <mdb.h>
#include <nmap.h>
#include <nmlib.h>
#include <msgapi.h>
#include <libical.h>
#include <streamio.h>
#include <openssl/md5.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <connio.h>

#include "calcmdd.h"


#if defined(RELEASE_BUILD)
#define CalCmdClientAlloc() MemPrivatePoolGetEntry(CalCmd.nmap.pool)
#else
#define CalCmdClientAlloc() MemPrivatePoolGetEntryDebug(CalCmd.nmap.pool, __FILE__, __LINE__)
#endif

static void SignalHandler(int sigtype);

#define QUEUE_WORK_TO_DO(c, id, r) \
        { \
            XplWaitOnLocalSemaphore(CalCmd.nmap.semaphore); \
            if (XplSafeRead(CalCmd.nmap.worker.idle)) { \
                (c)->queue.previous = NULL; \
                if (((c)->queue.next = CalCmd.nmap.worker.head) != NULL) { \
                    (c)->queue.next->queue.previous = (c); \
                } else { \
                    CalCmd.nmap.worker.tail = (c); \
                } \
                CalCmd.nmap.worker.head = (c); \
                (r) = 0; \
            } else { \
                XplSafeIncrement(CalCmd.nmap.worker.active); \
                XplSignalBlock(); \
                XplBeginThread(&(id), HandleConnection, 24 * 1024, XPL_INT_TO_PTR(XplSafeRead(CalCmd.nmap.worker.active)), (r)); \
                XplSignalHandler(SignalHandler); \
                if (!(r)) { \
                    (c)->queue.previous = NULL; \
                    if (((c)->queue.next = CalCmd.nmap.worker.head) != NULL) { \
                        (c)->queue.next->queue.previous = (c); \
                    } else { \
                        CalCmd.nmap.worker.tail = (c); \
                    } \
                    CalCmd.nmap.worker.head = (c); \
                } else { \
                    XplSafeDecrement(CalCmd.nmap.worker.active); \
                    (r) = -1; \
                } \
            } \
            XplSignalLocalSemaphore(CalCmd.nmap.semaphore); \
        }


CalCmdGlobals CalCmd;

static BOOL 
CalCmdClientAllocCB(void *buffer, void *data)
{
    register CalCmdClient *c = (CalCmdClient *)buffer;

    memset(c, 0, sizeof(CalCmdClient));

    return(TRUE);
}

static void 
CalCmdClientFree(CalCmdClient *client)
{
    register CalCmdClient *c = client;

    if (c->conn) {
        ConnClose(c->conn, 1);
        ConnFree(c->conn);
        c->conn = NULL;
    }

    MemPrivatePoolReturnEntry(c);

    return;
}

static void 
FreeClientData(CalCmdClient *client)
{
    if (client && !(client->flags & CALCMD_CLIENT_FLAG_EXITING)) {
        client->flags |= CALCMD_CLIENT_FLAG_EXITING;

        if (client->conn) {
            ConnClose(client->conn, 1);
            ConnFree(client->conn);
            client->conn = NULL;
        }

        if (client->envelope) {
            MemFree(client->envelope);
        }
    }

    return;
}





/* a quick 'n dirty urlencoder.  someone should write an optimized one 
   and put it in src/libs/ */

static char
numToHex[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 
                 'a', 'b', 'c', 'd', 'e', 'f' };

/* buffer must be at least 3 times the size of string */

static char *
URLEncode(char *buffer, char *string)
{
    char *p = buffer;
    char *s = string;

    while (*s) {
        if (isalpha(*s) || isdigit(*s)) {
            *p++ = *s++;
        } else {
            *p++ = '%';
            *p++ = numToHex[*s / 16];
            *p++ = numToHex[*s++ % 16];
        }
    }
    *p = 0;
    return buffer;
}



/* returns num chars printed on success, -1 on failure */

static int
QueryCalendar(CalCmdClient *client, CalCmdCommand *cmd, char *buffer, size_t size)
{
    int ccode = 0;
    Connection *qconn = NULL;
    char qbuf[256];
    char buf[1024];
    char buf2[1024];
    char *ptr;
    int len = 0;
    struct tm begin;
    struct tm end;

    /* connect to Lucene server at localhost:8081 */

    qconn = ConnAlloc(TRUE);
    if (!qconn) {
        goto query_abort;
    }
    qconn->socketAddress.sin_family = AF_INET;
    qconn->socketAddress.sin_port = htons(8081);
    qconn->socketAddress.sin_addr.s_addr = inet_addr("127.0.0.1");

    XplConsolePrintf("CALCMD connecting to lucene server\r\n");

    if (-1 == ConnConnectEx(qconn, NULL, 0, NULL, client->conn->trace.destination)) {
        goto query_abort;
    }

    CalCmdLocalToUtc(&cmd->begin, &begin);
    CalCmdLocalToUtc(&cmd->end, &end);

    if (CALCMD_QUERY_CONFLICTS == cmd->type) {
        struct tm weekago;
        struct tm weeklater;

        memcpy(&weekago, &begin, sizeof(weekago));
        weekago.tm_mday -= 7;
        mktime(&weekago);

        memcpy(&weeklater, &end, sizeof(weeklater));
        weeklater.tm_mday += 7;
        mktime(&weeklater);

        snprintf(qbuf, sizeof(qbuf),
                 "hula.store:calendar AND ical.type:VEVENT AND "
                 "hula.start-time:[%04d%02d%02d%02d%02d TO %04d%02d%02d%02d%02d] "
                 "AND hula.end-time:[%04d%02d%02d%02d%02d TO %04d%02d%02d%02d%02d]",
                 weekago.tm_year + 1900, weekago.tm_mon + 1, weekago.tm_mday,
                 weekago.tm_hour, weekago.tm_min,
                 end.tm_year + 1900, end.tm_mon + 1, end.tm_mday,
                 end.tm_hour, end.tm_min,
                 begin.tm_year + 1900, begin.tm_mon + 1, begin.tm_mday,
                 begin.tm_hour, begin.tm_min,
                 weeklater.tm_year + 1900, weeklater.tm_mon + 1, weeklater.tm_mday,
                 weeklater.tm_hour, weeklater.tm_min);
    } else {
        snprintf(qbuf, sizeof(qbuf), 
                 "hula.store:calendar AND ical.type:VEVENT AND "
                 "hula.start-time:[%04d%02d%02d%02d TO %04d%02d%02d%02d]",
                 begin.tm_year + 1900, begin.tm_mon + 1, begin.tm_mday,
                 begin.tm_hour,
                 end.tm_year + 1900, end.tm_mon + 1, end.tm_mday,
                 end.tm_hour);
    }
    

    XplConsolePrintf ("CALCMD sending query %s\r\n", qbuf);

    ccode = ConnWriteF(qconn, 
                       "GET /search/%s?query=%s&sortkey=hula.start-time"
                       "&sortorder=ascending&format=cal HTTP/1.0\r\n\r\n",
                       URLEncode(buf, client->user), URLEncode(buf2, qbuf));
    if (ccode <= 0) {
        goto query_abort;
    }
    ConnFlush (qconn);

    XplConsolePrintf ("CALCMD reading response \r\n");

    do { /* skip httpd headers*/
        ccode = ConnReadLine(qconn, client->line, CONN_BUFSIZE);
        if (ccode <= 0) {
            goto query_abort;
        }
        XplConsolePrintf ("CALCMD read header %s", client->line);
    } while (*client->line && !isspace(*client->line));

    do {
        struct tm begin;
        struct tm end;

        ccode = ConnReadLine(qconn, client->line, CONN_BUFSIZE);
        if (ccode < 0) {
            break;
        }
        client->line[CONN_BUFSIZE] = 0;
        XplConsolePrintf ("%s", client->line);
        
        if (' ' == *client->line) { /* continuation of previous appt summary */
            len += snprintf(&buffer[len], size - len, "%s", client->line + 1);
            continue;
        }

        /* line has format "(guid)|YYYYMMDDHHMM|YYYYMMDDHHMM|SUMMARY:(text)" */

        ptr = strchr(client->line, '|');
        if (!ptr) {
            goto query_abort;
        } 
        sscanf(ptr + 1, 
               "%4d%2d%2d%2d%2d|%4d%2d%2d%2d%2d|SUMMARY:",
               &begin.tm_year, &begin.tm_mon, &begin.tm_mday, 
               &begin.tm_hour, &begin.tm_min,
               
               &end.tm_year, &end.tm_mon, &end.tm_mday, 
               &end.tm_hour, &end.tm_min);

        begin.tm_year -= 1900;
        begin.tm_mon--;
        begin.tm_sec = 0;
        CalCmdUtcToLocal(&begin, &begin);

        end.tm_year -= 1900;
        end.tm_mon--;            
        end.tm_sec = 0;
        CalCmdUtcToLocal(&end, &end);
        
        ptr = strchr(ptr, ':');
        if (!ptr) {
            break;
        }
        ptr++;

        if (0 == len && CALCMD_QUERY_CONFLICTS == cmd->type) {
            len += snprintf(buffer, size - len, "Overlap found:\r\n");
        }

        len += CalCmdPrintConciseDateTimeRangeTM(&begin, &end, 
                                                 &buffer[len], size - len);
        if ((unsigned int)len >= size) {
            break;
        }
        len += snprintf(&buffer[len], size - len, " %s", ptr);

    } while (ccode);
    buffer[size-1] = '0';

    ConnClose(qconn, 0);
    ConnFree(qconn);
    
    if (0 == len && CALCMD_QUERY_CONFLICTS != cmd->type) {
        len += snprintf(buffer, size, "No events found.\r\n");
    }

    return len;

query_abort:
    XplConsolePrintf ("CALCMD QueryCalendar aborting %d\r\n", ccode);
    if (qconn) {
        ConnClose(qconn, 0);
        ConnFree(qconn);
    }
    return -1;  
}



/* returns non-zero on success */

static int
AddEventToCalendar(CalCmdClient *client, CalCmdCommand *calcmd)
{
    int ccode = 0;
    char icalbuf[4000];
    int len;
    char guid[NMAP_GUID_LENGTH+1];
    
    XplConsolePrintf ("CALCMD AddEventToCalendar\r\n");

    ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE, 
                            "GUID\r\n");
    if (1000 != ccode) {
        goto add_event_abort;
    }
    strncpy(guid, client->line, NMAP_GUID_LENGTH);
    guid[NMAP_GUID_LENGTH] = 0;

    len = CalCmdProcessNewAppt (calcmd, icalbuf, sizeof(icalbuf), guid);
    if (0 == len) {
        goto add_event_abort;
    }

    XplConsolePrintf ("CALCMD creating q entry\r\n");

    ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                            "QCREA\r\n");
    if (1000 != ccode) {
        goto add_event_abort; 
    }

    XplConsolePrintf ("CALCMD issuing qstor from command\r\n");

    ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                            "QSTOR FROM calcmdd -\r\n");
    if (1000 != ccode) {
        goto add_event_abort; 
    }


    XplConsolePrintf ("CALCMD issuing qstor cal command\r\n");

    ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                            "QSTOR CAL %s Personal\r\n", client->recipient);
    if (1000 != ccode) {
        goto add_event_abort; 
    }

    XplConsolePrintf ("CALCMD storing appt (len %d):\r\n%s\r\n", 
                      len, icalbuf);

    ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                            "QSTOR MESSAGE %d\r\n%s", len, icalbuf);
    if (1000 != ccode) {
        goto add_event_abort; 
    }
    XplConsolePrintf ("CALCMD putting cal obj in queue\r\n");

    ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                            "QRUN\r\n");
    if (1000 != ccode) {
        goto add_event_abort; 
    }

    XplConsolePrintf ("CALCMD it worked!\r\n");
    
    return 1;  
    
add_event_abort:
    XplConsolePrintf ("CALCMD AddEventToCalendar aborting %d %s\r\n", 
                      ccode, client->line);
    return 0;  
}








static __inline int 
ProcessConnection(CalCmdClient *client)
{
    int ccode;
    int length;
    int offset = 0;
    char *ptr;
    char *ptr2;
    char qID[16];
    char *cur;
    char *line;
    size_t size;
    CalCmdCommand calcmd;
    MDBValueStruct *vs;
    char *suffixCode;
    

    XplConsolePrintf ("CALCMD ProcessConnection\r\n");
    
    if (((ccode = NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, TRUE)) != -1) 
            && (ccode == 6020) 
            && ((ptr = strchr(client->line, ' ')) != NULL)) {
        *ptr++ = '\0';

        strcpy(qID, client->line);

        length = atoi(ptr);
        client->envelope = MemMalloc(length + 3);
    } else {
        NMAPSendCommand(client->conn, "QDONE\r\n", 7);
        NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, FALSE);
        return(-1);
    }

    if (client->envelope) {
        sprintf(client->line, "CalCmd: %s", qID);
        XplRenameThread(XplGetThreadID(), client->line);

        ccode = ConnRead(client->conn, client->envelope, length);
    } else {
        NMAPSendCommand(client->conn, "QDONE\r\n", 7);
        NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, FALSE);
        return(-1);
    }

    if ((ccode != -1) 
        && ((ccode = NMAPReadAnswer(client->conn, client->line, 
                                    CONN_BUFSIZE, TRUE)) == 6021)) 
    {
        client->envelope[length] = '\0';

    } else {
        goto process_conn_abort;
    }

    for (cur = client->envelope; *cur; cur = line) {
        line = strchr(cur, 0x0A);
        if (line) {
            if (line[-1] == 0x0D) {
                line[-1] = '\0';
            } else {
                *line = '\0';
            }
            line++;
        } else {
            line = cur + strlen(cur);
        }
        
        XplConsolePrintf ("CALCMD line: %s\r\n", cur);

        if (QUEUE_FROM == *cur) { 
            /* remember the sender */
            ptr = strchr(cur, ' ');
            *ptr = '\0';
            strcpy(client->sender, cur + 1);
            continue;
        }

        /* we're looking for local delivery to address of form "user+code" */
        if (QUEUE_RECIP_LOCAL != *cur || !(ptr = strchr(cur, '+'))) {
            continue;
        }
        *ptr = '\0';
        strcpy(client->recipient, cur + 1);        
        *ptr = '+';
        ptr2 = strchr(cur, ' ');
        if (ptr2) {
            *ptr2 = '\0';
        }

        /* lookup the user's secret suffix code */

        vs = MDBCreateValueStruct(CalCmd.handle.directory, MsgGetServerDN(NULL));
        if (!vs) {
            goto process_conn_abort;
        }
        if (!MsgFindObject(client->recipient, client->dn, NULL, NULL, vs)) {
            continue;
        }       
        strcpy(client->user, vs->Value[0]);
        MDBFreeValues(vs);

        XplConsolePrintf ("looking up: %s %s\r\n", 
                          client->dn, MSGSRV_A_CALCMD_ADDRESS_SUFFIX);
        ccode = MDBRead (client->dn, MSGSRV_A_CALCMD_ADDRESS_SUFFIX, vs);
        if (0 == ccode) {  /* HACK: until we have a form to 
                            * change your suffix, everyone will
                            * default to "123" */
            suffixCode = "123";
        } else {
            suffixCode = vs->Value[0];
        }
        ccode = strcmp(ptr + 1, suffixCode);

        XplConsolePrintf ("sekrit code: %s %lu\r\n", suffixCode, vs->ErrNo);

        MDBFreeValues(vs);
        if (ccode) {
            /* wrong secret code */
            continue;
        }

        /* extract MIME info; we only process only the first text/plain part */

        length = 0;
        ccode = NMAPSendCommandF(client->conn, "QMIME %s\r\n", qID);
        if (-1 == ccode) {
            goto process_conn_abort;
        }
        do {
            NmapMimeInfo minfo;

            ccode = NMAPReadResponse(client->conn, client->line, 
                                     CONN_BUFSIZE, TRUE);
            XplConsolePrintf ("qmime line %d %s\r\n", ccode, client->line);

            switch (ccode) {
            case 1000: /* done */
                break;
            case 2002:
                if (!NMAPParseMIMELine(client->line, &minfo)) {
                    goto process_conn_abort;
                }
                if (length || /* if length != 0, got one already, so ignore rest */
                    XplStrCaseCmp(minfo.type, "text") ||
                    XplStrCaseCmp(minfo.subtype, "plain")) 
                {
                    continue;
                }
                offset = minfo.start;
                length = minfo.size;
                continue;
            case 2003: /* multipart end */
            case 2004: /* rfc822 end */
                continue;
            default:
            case -1:
                goto process_conn_abort;
            }
        } while (1000 != ccode);

        if (length > CONN_BUFSIZE) {
            length = CONN_BUFSIZE;
        }
        ccode = NMAPSendCommandF(client->conn, "QBRAW %s %d %d\r\n", 
                                 qID, offset, length);
        if (-1 == ccode) {
            goto process_conn_abort;
        }
        
        ccode = NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, TRUE);
        if (2021 == ccode) {
            size = atoi (client->line);
            if (size > CONN_BUFSIZE) {
                size = CONN_BUFSIZE;
            }
        } else if (2025 == ccode) {
            size = length;
        } else {
            continue;
        }
        
        ccode = NMAPRead(client->conn, client->calreq, size);
        if (-1 == ccode) {
            goto process_conn_abort;
        }
        ccode = NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, TRUE);
        if (1000 != ccode) {
            goto process_conn_abort;
        }

        client->calreq[size] = 0;
        XplConsolePrintf ("CALCMD body %d:\r\n\%s\r\n", (int) size, client->calreq);
        
        {
            char reply[1024];
            int rlen = 0;

            char conflicts[512];
            int nlen;

            size = sizeof(reply);

            if (!CalCmdParseCommand(client->calreq, &calcmd)) {
                calcmd.type = CALCMD_ERROR;
            } 

            switch (calcmd.type) {
            case CALCMD_QUERY_CONFLICTS:
            case CALCMD_QUERY: 
                rlen += snprintf (&reply[rlen], size - rlen, 
                                  "Subject: results\r\n"
                                  "To: %s\r\n\r\n",
                                  client->sender);
                ccode = QueryCalendar(client, &calcmd, &reply[rlen], size - rlen);
                if (-1 == ccode) {
                    rlen = 0;
                    rlen += snprintf (&reply[rlen], size - rlen, 
                                      "Subject: calendar error\r\n"
                                      "To: %s\r\n\r\n",
                                      client->sender);
                    rlen += snprintf(&reply[rlen], size - rlen, 
                                     "Sorry, error querying calendar.\r\n");
                    reply[size-1] = '\0';
                } else {
                    rlen += ccode;
                }
                break;
            case CALCMD_NEW_APPT:
                calcmd.type = CALCMD_QUERY_CONFLICTS;
                nlen = QueryCalendar(client, &calcmd, conflicts, 
                                     sizeof(conflicts));
                calcmd.type = CALCMD_NEW_APPT;
                

                XplConsolePrintf ("CALCMD about to call AddEventToCalendar\r\n");
                
                if (!AddEventToCalendar(client, &calcmd)) {
                    rlen += snprintf (&reply[rlen], size - rlen, 
                                      "Subject: calendar error\r\n"
                                      "To: %s\r\n\r\n",
                                      client->sender);
                    rlen += snprintf(&reply[rlen], size - rlen, 
                                     "Sorry, unable to add event to calendar.\r\n");
                    reply[size-1] = '\0';
                } else {
                    rlen += snprintf (&reply[rlen], size - rlen, 
                                      "Subject: cal success\r\n"
                                      "To: %s\r\n\r\n",
                                      client->sender);
                    calcmd.data[90] = 0;
                    rlen += snprintf(&reply[rlen], size - rlen, 
                                     "OK, new appt: %s ", 
                                     calcmd.data);
                    rlen += 
                        CalCmdPrintConciseDateTimeRangeTM(&calcmd.begin, &calcmd.end,
                                                          &reply[rlen], size - rlen);
                    rlen += snprintf (&reply[rlen], size - rlen, "\r\n");

                    /* report any conflicts */
                    if (nlen) {
                        strncat(&reply[rlen], conflicts, size - rlen - 1);
                        rlen += nlen;
                    }
                    reply[size-1] = '\0';
                }
                break;
            case CALCMD_ERROR:
                rlen = 0;
                rlen += snprintf(reply, size - rlen, 
                                 "Subject: calendar error\r\n"
                                 "To: %s\r\n\r\n",
                                 client->sender);
                rlen += snprintf(&reply[rlen], size - rlen, 
                                 "Sorry, unable to process your message.  "
                                 "For help on Hula messages, please "
                                 "send the text 'help'.\r\n");
                reply[size-1] = '\0';
                break;
            case CALCMD_HELP:
                rlen = 0;
                rlen += snprintf(reply, size - rlen, 
                                 "Subject: calendar help\r\n"
                                 "To: %s\r\n\r\n",
                                 client->sender);
                rlen += snprintf(&reply[rlen], size - rlen, 
                                 "Send date/time to see your schedule, "
                                 "e.g. 'today' or 'fri nite'.\r\n"
                                 "Send date/time and summary to add a new "
                                 "item, e.g. 'meeting 3pm' or "
                                 "'Oct 15 payday'.\r\n");
                reply[size-1] = '\0';
                break;
            }

            ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                                    "QCREA 0\r\n");
            if (1000 != ccode) {
                goto process_conn_abort; 
            }
            XplConsolePrintf ("CALCMD issuing qstor from %s@%s %s@%s command\r\n",
                              client->recipient, CalCmd.officialName, 
                              client->recipient, CalCmd.officialName);

            ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                                    "QSTOR FROM %s@%s %s@%s\r\n",
                                    client->recipient, CalCmd.officialName, 
                                    client->recipient, CalCmd.officialName);

            if (1000 != ccode) {
                goto process_conn_abort; 
            }

            XplConsolePrintf ("CALCMD issuing qstor to %s %s command\r\n",
                              client->sender, client->sender);

            ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                                    "QSTOR TO %s %s 0\r\n", 
                                    client->sender, client->sender);
            if (1000 != ccode) {
                goto process_conn_abort; 
            }

            XplConsolePrintf ("CALCMD qstor message (len %d):\r\n%s\r\n", 
                              rlen, reply);

            ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                                    "QSTOR MESSAGE %d\r\n%s", rlen, reply);
            if (1000 != ccode) {
                goto process_conn_abort; 
            }

            ccode = NMAPRunCommandF(client->conn, client->line, CONN_BUFSIZE,
                                    "QRUN\r\n");
            if (1000 != ccode) {
                goto process_conn_abort; 
            }
        }

        /* no more further processing for this message */
        XplConsolePrintf ("CALCMD Deleting current q entry\r\n");
        ccode = NMAPSendCommandF(client->conn, "QDELE %s\r\n", qID);
        if (-1 == ccode) { 
            goto process_conn_abort; 
        }
        ccode = NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, TRUE);
        if (1000 != ccode) {
            goto process_conn_abort;
        }
        XplConsolePrintf ("CALCMD success!?!?  %s \r\n", client->line);
    }

    if ((ccode != -1) 
        && ((ccode = NMAPSendCommand(client->conn, "QDONE\r\n", 7)) != -1)) {
        ccode = NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, FALSE);
    }

    MemFree(client->envelope);
    client->envelope = NULL;


    return 0;

process_conn_abort:
    NMAPSendCommand(client->conn, "QDONE\r\n", 7);
    NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, FALSE);
    XplConsolePrintf ("CALCMD aborted %d %s\r\n", ccode, client->line);
    return -1;
}

static void 
HandleConnection(void *param)
{
    int ccode;
    long threadNumber = (long)param;
    time_t sleep = time(NULL);
    time_t wokeup;
    CalCmdClient *client;

    if ((client = CalCmdClientAlloc()) == NULL) {
        XplConsolePrintf("calcmdd: New worker failed to startup; out of memory.\r\n");

        NMAPSendCommand(client->conn, "QDONE\r\n", 7);

        XplSafeDecrement(CalCmd.nmap.worker.active);

        return;
    }

    do {
        XplRenameThread(XplGetThreadID(), "CalCmd Worker");

        XplSafeIncrement(CalCmd.nmap.worker.idle);

        XplWaitOnLocalSemaphore(CalCmd.nmap.worker.todo);

        XplSafeDecrement(CalCmd.nmap.worker.idle);

        wokeup = time(NULL);

        XplWaitOnLocalSemaphore(CalCmd.nmap.semaphore);

        client->conn = CalCmd.nmap.worker.tail;
        if (client->conn) {
            CalCmd.nmap.worker.tail = client->conn->queue.previous;
            if (CalCmd.nmap.worker.tail) {
                CalCmd.nmap.worker.tail->queue.next = NULL;
            } else {
                CalCmd.nmap.worker.head = NULL;
            }
        }

        XplSignalLocalSemaphore(CalCmd.nmap.semaphore);

        if (client->conn) {
            if (ConnNegotiate(client->conn, CalCmd.nmap.ssl.context)) {
                ccode = ProcessConnection(client);
            } else {
                NMAPSendCommand(client->conn, "QDONE\r\n", 7);
                NMAPReadAnswer(client->conn, client->line, CONN_BUFSIZE, FALSE);
            }
        }

        if (client->conn) {
            ConnFlush(client->conn);
        }

        FreeClientData(client);

        /* Live or die? */
        if (threadNumber == XplSafeRead(CalCmd.nmap.worker.active)) {
            if ((wokeup - sleep) > CalCmd.nmap.sleepTime) {
                break;
            }
        }

        sleep = time(NULL);

        CalCmdClientAllocCB(client, NULL);
    } while (CalCmd.state == CALCMD_STATE_RUNNING);

    FreeClientData(client);

    CalCmdClientFree(client);

    XplSafeDecrement(CalCmd.nmap.worker.active);

    XplExitThread(TSR_THREAD, 0);

    return;
}

static void 
CalCmdAgentServer(void *ignored)
{
    int i;
    int ccode;
    XplThreadID id;
    Connection *conn;

    XplSafeIncrement(CalCmd.server.active);

    XplRenameThread(XplGetThreadID(), "CalCmd Agent Server");

    CalCmd.state = CALCMD_STATE_RUNNING;

    while (CalCmd.state < CALCMD_STATE_STOPPING) {
        if (ConnAccept(CalCmd.nmap.conn, &conn) != -1) {
            if (CalCmd.state < CALCMD_STATE_STOPPING) {
                conn->ssl.enable = FALSE;

                QUEUE_WORK_TO_DO(conn, id, ccode);
                if (!ccode) {
                    XplSignalLocalSemaphore(CalCmd.nmap.worker.todo);

                    continue;
                }
            }

            ConnWrite(conn, "QDONE\r\n", 7);
            ConnClose(conn, 0);

            ConnFree(conn);
            conn = NULL;

            continue;
        }

        switch (errno) {
            case ECONNABORTED:
#ifdef EPROTO
            case EPROTO: 
#endif
            case EINTR: {
                if (CalCmd.state < CALCMD_STATE_STOPPING) {
                    LoggerEvent(CalCmd.handle.logging, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ERROR, 0, "Server", NULL, errno, 0, NULL, 0);
                }

                continue;
            }

            default: {
                if (CalCmd.state < CALCMD_STATE_STOPPING) {
                    XplConsolePrintf("calcmdd: Exiting after an accept() failure; error %d\r\n", errno);

                    LoggerEvent(CalCmd.handle.logging, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ERROR, 0, "Server", NULL, errno, 0, NULL, 0);

                    CalCmd.state = CALCMD_STATE_STOPPING;
                }

                break;
            }
        }

        break;
    }

    /* Shutting down */
    CalCmd.state = CALCMD_STATE_UNLOADING;

    XplConsolePrintf("calcmdd: Shutting down.\r\n");

    id = XplSetThreadGroupID(CalCmd.id.group);

    if (CalCmd.nmap.conn) {
        ConnClose(CalCmd.nmap.conn, 1);
        CalCmd.nmap.conn = NULL;
    }

    if (CalCmd.nmap.ssl.enable) {
        CalCmd.nmap.ssl.enable = FALSE;

        if (CalCmd.nmap.ssl.conn) {
            ConnClose(CalCmd.nmap.ssl.conn, 1);
            CalCmd.nmap.ssl.conn = NULL;
        }

        if (CalCmd.nmap.ssl.context) {
            ConnSSLContextFree(CalCmd.nmap.ssl.context);
            CalCmd.nmap.ssl.context = NULL;
        }
    }

    ConnCloseAll(1);

    if (ManagementState() == MANAGEMENT_RUNNING) {
        ManagementShutdown();
    }

    for (i = 0; (XplSafeRead(CalCmd.server.active) > 1) && (i < 60); i++) {
        XplDelay(1000);
    }

    for (i = 0; (ManagementState() != MANAGEMENT_STOPPED) && (i < 60); i++) {
        XplDelay(1000);
    }

    XplConsolePrintf("calcmdd: Shutting down %d queue threads\r\n", XplSafeRead(CalCmd.nmap.worker.active));

    XplWaitOnLocalSemaphore(CalCmd.nmap.semaphore);

    ccode = XplSafeRead(CalCmd.nmap.worker.idle);
    while (ccode--) {
        XplSignalLocalSemaphore(CalCmd.nmap.worker.todo);
    }

    XplSignalLocalSemaphore(CalCmd.nmap.semaphore);

    for (i = 0; XplSafeRead(CalCmd.nmap.worker.active) && (i < 60); i++) {
        XplDelay(1000);
    }

    if (XplSafeRead(CalCmd.server.active) > 1) {
        XplConsolePrintf("calcmdd: %d server threads outstanding; attempting forceful unload.\r\n", XplSafeRead(CalCmd.server.active) - 1);
    }

    if (XplSafeRead(CalCmd.nmap.worker.active)) {
        XplConsolePrintf("calcmdd: %d threads outstanding; attempting forceful unload.\r\n", XplSafeRead(CalCmd.nmap.worker.active));
    }

    LoggerClose(CalCmd.handle.logging);
    CalCmd.handle.logging = NULL;

    /* shutdown the scanning engine */

    XplCloseLocalSemaphore(CalCmd.nmap.worker.todo);
    XplCloseLocalSemaphore(CalCmd.nmap.semaphore);

    MsgShutdown();

    CONN_TRACE_SHUTDOWN();
    ConnShutdown();

    MemPrivatePoolFree(CalCmd.nmap.pool);

    MemoryManagerClose(MSGSRV_AGENT_CALCMD); 

    XplConsolePrintf("calcmdd: Shutdown complete\r\n");

    XplSignalLocalSemaphore(CalCmd.sem.main);
    XplWaitOnLocalSemaphore(CalCmd.sem.shutdown);

    XplCloseLocalSemaphore(CalCmd.sem.shutdown);
    XplCloseLocalSemaphore(CalCmd.sem.main);

    XplSetThreadGroupID(id);

    return;
}

static BOOL 
ReadConfiguration(void)
{
    MDBValueStruct *config;

    config = MDBCreateValueStruct(CalCmd.handle.directory, MsgGetServerDN(NULL));
    if (config) {
        if (MDBRead(".", MSGSRV_A_OFFICIAL_NAME, config)) {
            strcpy(CalCmd.officialName, config->Value[0]);

            MDBFreeValues(config);
        } else {
            CalCmd.officialName[0] = '\0';
        }

        MDBSetValueStructContext(NULL, config);
        if (MDBRead(MSGSRV_ROOT, MSGSRV_A_ACL, config) > 0) { 
            HashCredential(MsgGetServerDN(NULL), config->Value[0], CalCmd.nmap.hash);
        }
    } else {
        return(FALSE);
    }

    MDBDestroyValueStruct(config);

    return(TRUE);
}

#if defined(NETWARE) || defined(LIBC)
static int 
_NonAppCheckUnload(void)
{
    int            s;
    static BOOL    checked = FALSE;
    XplThreadID    id;

    if (!checked) {
        checked = TRUE;
        CalCmd.state = CALCMD_STATE_UNLOADING;

        XplWaitOnLocalSemaphore(CalCmd.sem.shutdown);

        id = XplSetThreadGroupID(CalCmd.id.group);
        ConnClose(CalCmd.nmap.conn, 1);
        XplSetThreadGroupID(id);

        XplWaitOnLocalSemaphore(CalCmd.sem.main);
    }

    return(0);
}
#endif

static void 
SignalHandler(int sigtype)
{
    switch(sigtype) {
        case SIGHUP: {
            if (CalCmd.state < CALCMD_STATE_UNLOADING) {
                CalCmd.state = CALCMD_STATE_UNLOADING;
            }

            break;
        }
        case SIGINT:
        case SIGTERM: {
            if (CalCmd.state == CALCMD_STATE_STOPPING) {
                kill(getpid(), 9);
            } else if (CalCmd.state < CALCMD_STATE_STOPPING) {
                CalCmd.state = CALCMD_STATE_STOPPING;
            }

            break;
        }

        default: {
            break;
        }
    }

    return;
}

static int 
QueueSocketInit(void)
{
    CalCmd.nmap.conn = ConnAlloc(FALSE);
    if (CalCmd.nmap.conn) {
        memset(&(CalCmd.nmap.conn->socketAddress), 0, sizeof(CalCmd.nmap.conn->socketAddress));

        CalCmd.nmap.conn->socketAddress.sin_family = AF_INET;
        CalCmd.nmap.conn->socketAddress.sin_addr.s_addr = MsgGetAgentBindIPAddress();

        /* Get root privs back for the bind.  It's ok if this fails -
        * the user might not need to be root to bind to the port */
        XplSetEffectiveUserId(0);

        CalCmd.nmap.conn->socket = ConnServerSocket(CalCmd.nmap.conn, 2048);
        if (XplSetEffectiveUser(MsgGetUnprivilegedUser()) < 0) {
            XplConsolePrintf("calcmdd: Could not drop to unprivileged user '%s'.\r\n", MsgGetUnprivilegedUser());
            ConnFree(CalCmd.nmap.conn);
            CalCmd.nmap.conn = NULL;
            return(-1);
        }

        if (CalCmd.nmap.conn->socket == -1) {
            XplConsolePrintf("calcmdd: Could not bind to dynamic port.\r\n");
            ConnFree(CalCmd.nmap.conn);
            CalCmd.nmap.conn = NULL;
            return(-1);
        }

        if (NMAPRegister(MSGSRV_AGENT_CALCMD, CalCmd.nmap.queue, CalCmd.nmap.conn->socketAddress.sin_port)
            != REGISTRATION_COMPLETED) 
        {
            XplConsolePrintf("calcmdd: Could not register with hulanmap.\r\n");
            ConnFree(CalCmd.nmap.conn);
            CalCmd.nmap.conn = NULL;
            return(-1);
        }

    } else {
        XplConsolePrintf("calcmdd: Could not allocate connection.\r\n");
        return(-1);
    }

    return(0);
}

XplServiceCode(SignalHandler)

int
XplServiceMain(int argc, char *argv[])
{
    int                ccode;
    XplThreadID        id;

    if (XplSetEffectiveUser(MsgGetUnprivilegedUser()) < 0) {
        XplConsolePrintf("calcmdd: Could not drop to unprivileged user '%s', exiting.\n", MsgGetUnprivilegedUser());
        return(1);
    }

    XplSignalHandler(SignalHandler);

    CalCmd.id.main = XplGetThreadID();
    CalCmd.id.group = XplGetThreadGroupID();

    CalCmd.state = CALCMD_STATE_INITIALIZING;
    CalCmd.flags = 0;

    CalCmd.nmap.conn = NULL;
    CalCmd.nmap.queue = Q_FOUR;
    CalCmd.nmap.pool = NULL;
    CalCmd.nmap.sleepTime = (5 * 60);
    CalCmd.nmap.ssl.conn = NULL;
    CalCmd.nmap.ssl.enable = FALSE;
    CalCmd.nmap.ssl.context = NULL;
    CalCmd.nmap.ssl.config.options = 0;

    CalCmd.handle.directory = NULL;
    CalCmd.handle.logging = NULL;

    strcpy(CalCmd.nmap.address, "127.0.0.1");

    XplSafeWrite(CalCmd.server.active, 0);

    XplSafeWrite(CalCmd.nmap.worker.idle, 0);
    XplSafeWrite(CalCmd.nmap.worker.active, 0);
    XplSafeWrite(CalCmd.nmap.worker.maximum, 100000);

    if (MemoryManagerOpen(MSGSRV_AGENT_CALCMD) == TRUE) {
        CalCmd.nmap.pool = MemPrivatePoolAlloc("CalCmdAgent Connections", sizeof(CalCmdClient), 0, 3072, TRUE, FALSE, CalCmdClientAllocCB, NULL, NULL);
        if (CalCmd.nmap.pool != NULL) {
            XplOpenLocalSemaphore(CalCmd.sem.main, 0);
            XplOpenLocalSemaphore(CalCmd.sem.shutdown, 1);
            XplOpenLocalSemaphore(CalCmd.nmap.semaphore, 1);
            XplOpenLocalSemaphore(CalCmd.nmap.worker.todo, 1);
        } else {
            MemoryManagerClose(MSGSRV_AGENT_CALCMD);

            XplConsolePrintf("calcmdd: Unable to create connection pool; shutting down.\r\n");
            return(-1);
        }
    } else {
        XplConsolePrintf("calcmdd: Unable to initialize memory manager; shutting down.\r\n");
        return(-1);
    }

    ConnStartup(CONNECTION_TIMEOUT, TRUE);

    MDBInit();
    CalCmd.handle.directory = (MDBHandle)MsgInit();
    if (CalCmd.handle.directory == NULL) {
        XplBell();
        XplConsolePrintf("calcmdd: Invalid directory credentials; exiting!\r\n");
        XplBell();

        MemoryManagerClose(MSGSRV_AGENT_CALCMD);

        return(-1);
    }

    NMAPInitialize(CalCmd.handle.directory);

    SetCurrentNameSpace(NWOS2_NAME_SPACE);
    SetTargetNameSpace(NWOS2_NAME_SPACE);

	CalCmd.handle.logging = LoggerOpen("hulacalcmd");
	if (!CalCmd.handle.logging) {
		XplConsolePrintf("calcmdd: Unable to initialize logging; disabled.\r\n");
	}

    ReadConfiguration();
    CONN_TRACE_INIT((char *)MsgGetWorkDir(NULL), "calcmd");
    // CONN_TRACE_SET_FLAGS(CONN_TRACE_ALL); /* uncomment this line and pass '--enable-conntrace' to autogen to get the agent to trace all connections */

    if (QueueSocketInit() < 0) {
        XplConsolePrintf("calcmdd: Exiting.\r\n");

        MemoryManagerClose(MSGSRV_AGENT_CALCMD);

        return -1;
    }

    /* initialize scanning engine here */


    if (XplSetRealUser(MsgGetUnprivilegedUser()) < 0) {
        XplConsolePrintf("calcmdd: Could not drop to unprivileged user '%s', exiting.\r\n", MsgGetUnprivilegedUser());

        MemoryManagerClose(MSGSRV_AGENT_CALCMD);

        return 1;
    }

    CalCmd.nmap.ssl.enable = FALSE;
    CalCmd.nmap.ssl.config.method = SSLv23_client_method;
    CalCmd.nmap.ssl.config.options = SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS;
    CalCmd.nmap.ssl.config.mode = SSL_MODE_AUTO_RETRY;
    CalCmd.nmap.ssl.config.cipherList = NULL;
    CalCmd.nmap.ssl.config.certificate.type = SSL_FILETYPE_PEM;
    CalCmd.nmap.ssl.config.certificate.file = MsgGetTLSCertPath(NULL);
    CalCmd.nmap.ssl.config.key.type = SSL_FILETYPE_PEM;
    CalCmd.nmap.ssl.config.key.file = MsgGetTLSKeyPath(NULL);

    CalCmd.nmap.ssl.context = ConnSSLContextAlloc(&(CalCmd.nmap.ssl.config));
    if (CalCmd.nmap.ssl.context) {
        CalCmd.nmap.ssl.enable = TRUE;
    }

    NMAPSetEncryption(CalCmd.nmap.ssl.context);

    ccode = -1;
    if ((ManagementInit(MSGSRV_AGENT_CALCMD, CalCmd.handle.directory)) 
            && (ManagementSetVariables(GetCalCmdManagementVariables(), GetCalCmdManagementVariablesCount())) 
            && (ManagementSetCommands(GetCalCmdManagementCommands(), GetCalCmdManagementCommandsCount()))) {
        XplBeginThread(&id, ManagementServer, DMC_MANAGEMENT_STACKSIZE, NULL, ccode);
    }


    if (ccode) {
        XplConsolePrintf("calcmdd: Unable to startup the management interface.\r\n");
    }
    
    XplStartMainThread(PRODUCT_SHORT_NAME, &id, CalCmdAgentServer, 8192, NULL, ccode);
    
    XplUnloadApp(XplGetThreadID());
    return(0);
}
