/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include <xpl.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <stdio.h>
#include <nmap.h>
#include <hulautil.h>
#include <nmlib.h>

#include "search.h"
#include "store.h"
#include "fetch.h"
#include "rfc822parse.h"

/* Port to listen */
#define IMAP4_PORT  143
#define IMAP4_PORT_SSL 993

/* return values for command functions */
typedef enum {
    STATUS_ABORT = -1,
    STATUS_CONTINUE = 0,
    STATUS_IDENTITY_STORE_FAILURE,
    STATUS_USERNAME_NOT_FOUND,
    STATUS_USERNAME_TOO_LONG,
    STATUS_WRONG_PASSWORD,
    STATUS_UNKNOWN_AUTH,
    STATUS_NMAP_COMM_ERROR,
    STATUS_NMAP_CONNECT_ERROR,
    STATUS_NMAP_AUTH_ERROR,
    STATUS_FEATURE_DISABLED,
    STATUS_INVALID_FOLDER_NAME,
    STATUS_NO_MATCHING_FOLDER,
    STATUS_READ_ONLY_FOLDER,
    STATUS_COULD_NOT_DETERMINE_QUOTA,
    STATUS_NO_SUCH_QUOTA_ROOT,
    STATUS_MEMORY_ERROR,
    STATUS_PERMISSION_DENIED,
    STATUS_TRY_CREATE,
    STATUS_INVALID_ARGUMENT,
    STATUS_INVALID_STATE,
    STATUS_UNKNOWN_COMMAND,
    STATUS_TLS_NEGOTIATION_FAILURE,
    STATUS_CHARSET_NOT_SUPPORTED,
    STATUS_UID_NOT_FOUND,
    STATUS_NMAPID_NOT_FOUND,
    STATUS_IMAPID_NOT_FOUND,

    STATUS_MAX
} CommandStatus;

typedef struct {
    long id;
    char *formatString;
} ImapErrorString;

/* STATUS_USERNAME_NOT_FOUND, STATUS_WRONG_PASSWORD, and STATUS_PERMISSION_DENIED */
/* have the same string to prevent spammers from mining the user namespace */

static ImapErrorString ImapErrorStrings[] = {
    { STATUS_IDENTITY_STORE_FAILURE,    "%s NO %s the directory used to lookup users is not functional\r\n" },
    { STATUS_USERNAME_NOT_FOUND,        "%s NO %s permission denied\r\n" },
    { STATUS_USERNAME_TOO_LONG,         "%s NO %s username or its alias exceeds system limits\r\n" },
    { STATUS_WRONG_PASSWORD,            "%s NO %s permission denied\r\n" },
    { STATUS_UNKNOWN_AUTH,              "%s NO %s mechanism unknown\r\n" },
    { STATUS_NMAP_COMM_ERROR,           "%s NO %s unable to communicate with the mail store\r\n" },
    { STATUS_NMAP_CONNECT_ERROR,        "%s NO %s unable to connect to the mail store\r\n" },
    { STATUS_NMAP_AUTH_ERROR,           "%s NO %s unable to authenticate to the mail store\r\n" },
    { STATUS_FEATURE_DISABLED,          "%s NO %s the IMAP service is not enabled for this user\r\n" },
    { STATUS_INVALID_FOLDER_NAME,       "%s NO %s illegal folder name\r\n" },
    { STATUS_NO_MATCHING_FOLDER,        "%s NO %s pattern doesn't match mailbox\r\n" },
    { STATUS_READ_ONLY_FOLDER,          "%s NO %s read only folder! Your IMAP client ignored the response to the select command\r\n" },
    { STATUS_COULD_NOT_DETERMINE_QUOTA, "%s NO %s quota could not be determined\r\n" },
    { STATUS_NO_SUCH_QUOTA_ROOT,        "%s NO %s no such quota root\r\n" },
    { STATUS_MEMORY_ERROR,              "%s NO %s out of memory\r\n" },
    { STATUS_PERMISSION_DENIED,         "%s NO %s permission denied\r\n" },
    { STATUS_TRY_CREATE,                "%s NO %s [TRYCREATE]\r\n" },
    { STATUS_CHARSET_NOT_SUPPORTED,     "%s NO %s charset not supported\r\n" },
    { STATUS_INVALID_ARGUMENT,          "%s BAD %s invalid arguments\r\n" },
    { STATUS_INVALID_STATE,             "%s BAD %s invalid command in current state\r\n" },
    { STATUS_UNKNOWN_COMMAND,           "* BAD command unrecognized\r\n" },
    { STATUS_TLS_NEGOTIATION_FAILURE,   "%s BAD %s negotiation failed\r\n" },
    { 0, NULL }
};

/* supported IMAP commands */

/* IMAP client commands - any state */
#define IMAP_COMMAND_CAPABILITY "CAPABILITY"
#define IMAP_COMMAND_NOOP "NOOP"
#define IMAP_COMMAND_LOGOUT "LOGOUT"

/* IMAP client commands - not authenticated state */
#define IMAP_COMMAND_STARTTLS "STARTTLS"
#define IMAP_COMMAND_AUTHENTICATE "AUTHENTICATE"
#define IMAP_COMMAND_LOGIN "LOGIN"

/* IMAP client commands - authenticated state */
#define IMAP_COMMAND_SELECT "SELECT"
#define IMAP_COMMAND_EXAMINE "EXAMINE"
#define IMAP_COMMAND_CREATE "CREATE"
#define IMAP_COMMAND_DELETE "DELETE"
#define IMAP_COMMAND_RENAME "RENAME"
#define IMAP_COMMAND_SUBSCRIBE "SUBSCRIBE"
#define IMAP_COMMAND_UNSUBSCRIBE "UNSUBSCRIBE"
#define IMAP_COMMAND_LIST "LIST"
#define IMAP_COMMAND_LSUB "LSUB"
#define IMAP_COMMAND_STATUS "STATUS"
#define IMAP_COMMAND_APPEND "APPEND"

/* IMAP client commands - selected state */
#define IMAP_COMMAND_CHECK "CHECK"
#define IMAP_COMMAND_CLOSE "CLOSE"
#define IMAP_COMMAND_EXPUNGE "EXPUNGE"
#define IMAP_COMMAND_SEARCH "SEARCH"
#define IMAP_COMMAND_UID_SEARCH "UID SEARCH"
#define IMAP_COMMAND_FETCH "FETCH"
#define IMAP_COMMAND_UID_FETCH "UID FETCH"
#define IMAP_COMMAND_STORE "STORE"
#define IMAP_COMMAND_UID_STORE "UID STORE"
#define IMAP_COMMAND_COPY "COPY"
#define IMAP_COMMAND_UID_COPY "UID COPY"

/* IMAP ACL commands rfc2086 */
#define IMAP_COMMAND_SETACL "SETACL"
#define IMAP_COMMAND_DELETEACL "DELETEACL"
#define IMAP_COMMAND_GETACL "GETACL"
#define IMAP_COMMAND_LISTRIGHTS "LISTRIGHTS"
#define IMAP_COMMAND_MYRIGHTS "MYRIGHTS"

/* IMAP Quota commands rfc2087 */
#define IMAP_COMMAND_SETQUOTA "SETQUOTA"
#define IMAP_COMMAND_GETQUOTA "GETQUOTA"
#define IMAP_COMMAND_GETQUOTAROOT "GETQUOTAROOT"

/* IMAP Namespace command rfc2342 */
#define IMAP_COMMAND_NAMESPACE "NAMESPACE"

/* non-standard commands */
#define IMAP_COMMAND_PROXYAUTH "PROXYAUTH"  /* depricated - should use SASL */

#define IMAP_HELP_NOT_DEFINED "%s - HELP not defined.\r\n"


/* Internal stuff */
#define CR     0x0d
#define LF     0x0a

/* Command State */
#define STATE_FRESH  (1<<0)
#define STATE_AUTH  (1<<1)
#define STATE_SELECTED (1<<2)
#define STATE_EXITING (1<<3)

#define BUFSIZE    1023

/* Minutes * 60 (1 second granularity) */
#define CONNECTION_TIMEOUT (40*60)

#define UID_HIGHEST     ULONG_MAX

#define IMAP_STACK_SIZE  (1024*65)


typedef struct {
    Connection            *conn;
    Connection            *nmap;
    int                   state;                /* Connection state    */
    unsigned char         MBox[XPL_MAX_PATH];   /* Greeting (for APOP)   */
    unsigned char         User[65];             /* Username       */
    unsigned char         *Command;             /* Current command    */
    unsigned long         CommandLen;           /* Allocated size of command */
    unsigned char         Ident[65];            /* Ident for current command */
    int                   LoginCount;           /* Internal       */
    unsigned long         Messages;             /* Number of messages loaded */
    unsigned long         NMAPCount;            /* Total number of messages   */
    unsigned long         *UIDList;             /* UID Cache      */
    unsigned long         *IDList;              /* ID Cache       */
    BOOL                  ReadOnly;             /* NMAP MBOX is read-only  */
    IMAPSEARCH            *Search;
    unsigned long         NMAPAsyncEvent;       /* Async Events received from NMAP */
    unsigned long        *UnseenFlags;          /* Used by HandleConnection() */
    unsigned char         dn[MDB_MAX_OBJECT_CHARS + 1];
} ImapClient;

typedef struct {
    struct {
        HulaKeywordIndex *index;
        long *returnValueIndex;

        char *months[12];

        struct {
            int len;
            char message[128];

            struct {
                int len;
                char message[128];
            } ssl;

            struct {
                BOOL enabled;
            } acl;
        } capability;

        struct{
            HulaKeywordIndex *optionIndex;
        } status;

        struct {
            HulaKeywordIndex *typeIndex;
            HulaKeywordIndex *flagIndex;
        } store;
    } command;

    struct {
        MDBHandle handle;
    } directory;

    struct {
        struct {
            XplAtomic inUse;
            XplAtomic idle;
            unsigned long max;
        } threads;
        
        XplAtomic served;
        XplAtomic badPassword;
    } client;

    struct {
        struct {
            BOOL enable;

            ConnSSLConfiguration config;

            SSL_CTX *context;

            Connection *conn;
        
            unsigned short port;
        } ssl;

        Connection *conn;

        XplAtomic active;

        struct sockaddr_in addr;

        char hostname[256];
        unsigned long  hostnameLen;
        unsigned short port;
        int threadGroupId;
        BOOL disabled;
        XplSemaphore mainSemaphore;
        XplSemaphore shutdownSemaphore;
        char postmaster[MAXEMAILNAMESIZE + 1];
    } server;

    void *logHandle;
    BOOL exiting;
} ImapGlobal;

extern ImapGlobal Imap;

/* IMAP prototypes */
long UIDtoNUM(ImapClient *Client, unsigned long *UID);
long UIDtoNUMRange(ImapClient *Client, unsigned long StartUID, unsigned long EndUID, long *StartIndex, long *EndIndex);
BOOL NmapAsyncCB(void *param, char *beginPtr, char *endPtr);
long NewMessageNotify(ImapClient *client);
long GrabArgument(ImapClient *client, unsigned char **Input, unsigned char **Destination);
long LoadUIDList(ImapClient *client);

void CommandFetchCleanup(void);
BOOL CommandFetchInit(void);
void CommandStoreCleanup(void);
BOOL CommandStoreInit(void);
void CommandStatusCleanup(void);
BOOL CommandStatusInit(void);

/* IMAP command prototypes */
int ImapCommandCapability(void *param);
int ImapCommandNoop(void *param);
int ImapCommandLogout(void *param);
int ImapCommandStartTls(void *param);
int ImapCommandAuthenticate(void *param);
int ImapCommandLogin(void *param);
int ImapCommandSelect(void *param);
int ImapCommandExamine(void *param);
int ImapCommandCreate(void *param);
int ImapCommandDelete(void *param);
int ImapCommandRename(void *param);
int ImapCommandSubscribe(void *param);
int ImapCommandUnsubscribe(void *param);
int ImapCommandList(void *param);
int ImapCommandLsub(void *param);
int ImapCommandStatus(void *param);
int ImapCommandAppend(void *param);
int ImapCommandCheck(void *param);
int ImapCommandClose(void *param);
int ImapCommandExpunge(void *param);
int ImapCommandSearch(void *param);
int ImapCommandUidSearch(void *param);
int ImapCommandFetch(void *param);
int ImapCommandUidFetch(void *param);
int ImapCommandStore(void *param);
int ImapCommandUidStore(void *param);
int ImapCommandCopy(void *param);
int ImapCommandUidCopy(void *param);
int ImapCommandSetAcl(void *param);
int ImapCommandDeleteAcl(void *param);
int ImapCommandGetAcl(void *param);
int ImapCommandListRights(void *param);
int ImapCommandMyRights(void *param);
int ImapCommandSetQuota(void *param);
int ImapCommandGetQuota(void *param);
int ImapCommandGetQuotaRoot(void *param);
int ImapCommandNameSpace(void *param);
int ImapCommandProxyAuth(void *param);

#include "inline.h"
