/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2005 Novell, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail, you
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
****************************************************************************/
#include <config.h>
#include <xpl.h>
#include <logger.h>
#include <mdb.h>
#include <nmap.h>
#include <nmlib.h>

#include "nmapdp.h"

#define CONTROL_TABLE_CREATE "CREATE TABLE control (" \
   "id INTEGER PRIMARY KEY AUTOINCREMENT, " \
    "flags INTEGER DEFAULT 0, " \
    "version INTEGER NOT NULL, " \
    "storeId INTEGER NOT NULL, " \
    "storeSize INTEGER DEFAULT 0, " \
    "storeCount INTEGER DEFAULT 0, " \
    "modified DATETIME NOT NULL);"
#define CONTROL_READ "SELECT id, flags, version, storeId, storeSize, storeCount, julianday(modified) "\
   "FROM control " \
    "ORDER BY id;"
#define CONTROL_TABLE_DELETE "DROP TABLE control;"
#define CONTROL_UPDATE "UPDATE control " \
   "SET flags = ?, storeId = ?, storeSize = ?, storeCount = ?, modified = julianday('now') " \
    "WHERE id = ?;"
#define CONTROL_DELETE "DELETE FROM control WHERE id = ?;"
#define CONTROL_TABLE_INSERT "INSERT INTO control (" \
   "flags, version, storeId, storeSize, storeCount, modified) " \
    "VALUES (?, ?, ?, ?, ?, julianday('now'));"
#define CONTROL_TABLE_INSERT_INIT "INSERT INTO control (" \
   "version, storeId, modified) "\
    "VALUES (?, ?, julianday('now'));"

#define INDEX_TABLE_CREATE "CREATE TABLE contact (" \
   "id INTEGER PRIMARY KEY AUTOINCREMENT, " \
    "flags INTEGER DEFAULT 0, " \
    "guid VARCHARS(46) NOT NULL UNIQUE, " \
    "record INTEGER NOT NULL, " \
    "start INTEGER NOT NULL, " \
    "fn INTEGER NOT NULL, " \
    "n INTEGER NOT NULL, " \
    "size INTEGER NOT NULL, " \
    "indexed DATETIME DEFAULT 0, " \
    "modified DATETIME NOT NULL);"
#define INDEX_READ_ALL  "SELECT id, flags, guid, record, start, fn, n, size, julianday(indexed), julianday(modified) " \
   "FROM contact " \
    "ORDER BY id;"
#define INDEX_READ "SELECT id, flags, guid, record, start, fn, n, size, julianday(indexed), julianday(modified) " \
   "FROM contact " \
    "WHERE id = ?;"
#define INDEX_TABLE_DELETE "DROP TABLE contact;"
#define INDEX_DELETE_ALL "DELETE FROM contact;"
#define INDEX_DELETE "DELETE FROM contact WHERE id = ?;"
#define INDEX_INSERT "INSERT INTO contact (" \
   "flags, guid, record, start, fn, n, size, indexed, modified) " \
    "VALUES (?, ?, ?, ?, ?, ?, ?, ?, julianday('now'));"
#define INDEX_UPDATE "UPDATE contact " \
   "SET flags = ?, record = ?, start = ?, fn = ?, n = ?, size = ?, indexed = ?, modified = julianday('now') " \
    "WHERE id = ?;"
#define INDEX_UPDATE_FLAGS "UPDATE contact " \
   "SET flags = ?, modified = julianday('now') " \
    "WHERE id = ?;"

static int OpenBook(NMAPClient *pClient);
static int CreateBook(NMAPClient *pClient);
static int IndexCreate(NMAPClient *pClient);
static int IndexOpen(NMAPClient *pClient);
static int IndexRead(NMAPClient *pClient);
static int ControlRead(NMAPClient *pClient, ContactControl *pControl);
static int ControlTableDelete(NMAPClient *pClient);
static int ControlTableCreate(NMAPClient *pClient, ContactControl *pControl);
static int ControlUpdate(NMAPClient *pClient);
static int IndexLoad(NMAPClient *pClient);
static int IndexTableDelete(NMAPClient *pClient);
static int IndexTableCreate(NMAPClient *pClient);
static int IndexBuild(NMAPClient *pClient);
static int IndexAdd(NMAPClient *pClient, ContactInfo *pContact);
static int IndexWrite(NMAPClient *pClient, ContactInfo *pContact);
static int TransBegin(NMAPClient *pClient);
static int TransCommit(NMAPClient *pClient);
static int TransAbort(NMAPClient *pClient);
static int Execute(NMAPClient *pClient, unsigned char *ptr);
static int Step(sqlite3_stmt *stmt);
static void FreeIndex(NMAPClient *pClient);
static int CheckBook(NMAPClient *pClient);
static int ParseVCard(
        NMAPClient *pClient,
        ContactInfo *pContact,
        unsigned char *pCard);
static int IndexUpdateFlags(NMAPClient *pClient, ContactInfo *pContact);

int
NmapCommandABook(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    unsigned char *ptr;
    int err;

    if (pClient->states & NMAP_CLIENT_USER) {
        if (!(pClient->states & NMAP_CLIENT_ABOOK)) {
            ptr = pClient->buffer + (sizeof(NMAP_ABOOK_COMMAND) - 1);
            if (*ptr != '\0') {
                pClient->states |= NMAP_CLIENT_ABOOK;
                memcpy(pBook->name, ++ptr, XPL_MAX_PATH);
                pBook->name[XPL_MAX_PATH + 1] = '\0';
                err = ReadNLockAquire(
                        &pBook->lock,
                        &pClient->userHash,
                        pClient->user,
                        pBook->name,
                        0);
                if (NLOCK_AQUIRED == err) {
                    err = OpenBook(pClient);
                    if (0 == err) {
                        if (pBook->cb.type & ContactReadOnly) {
                            (void) ConnWriteF(
                                    pClient->conn,
                                    "1020 %lu %s\r\n",
                                    pBook->control.storeId,
                                    MSG1020ADBK);
                        } else {
                            (void) ConnWriteF(
                                    pClient->conn,
                                    "1000 %lu %s\r\n",
                                    pBook->control.storeId,
                                    MSG1000ADBK);
                        }
                    }
                } else {
                    ReadNLockRelease(pBook->lock);
                    pBook->lock = NULL;
                }
            } else {
                err = ConnWrite(
                        pClient->conn,
                        MSG3010BADARGC,
                        sizeof(MSG3010BADARGC) - 1);
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG4221ADBKDONE,
                    sizeof(MSG4221ADBKDONE) - 1);
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG3241NOUSER,
                sizeof(MSG3241NOUSER) - 1);
    }
    return(err);
}

int
NmapCommandABCreate(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    unsigned char name[XPL_MAX_PATH + 1];
    unsigned char *ptr;
    int err;
    FILE *hFile;

    if (pClient->states & NMAP_CLIENT_USER) {
        ptr = pClient->buffer + (sizeof(NMAP_ABOOK_COMMAND) - 1);
        if (*ptr != '\0') {
            memcpy(name, ++ptr, XPL_MAX_PATH);
            name[XPL_MAX_PATH + 1] = '\0';
            sprintf(
                    pClient->path,
                    "%s/%s/%s"NMAP_ADBK_EXT,
                    pClient->store,
                    pClient->user,
                    name);
            hFile = fopen(pClient->path, "rb");
            if (hFile == NULL) {
                err = CreateBook(pClient);
                if (0 == err) {
                    hFile = fopen(pClient->path, "rb");
                    if (0 == hFile) {
                        err = ConnWrite(
                                pClient->conn,
                                MSG4240NOPERMISSION,
                                sizeof(MSG4240NOPERMISSION) - 1);
                    } else {
                        fclose(hFile);
                    }
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG3010BADARGC,
                    sizeof(MSG3010BADARGC) - 1);
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG3241NOUSER,
                sizeof(MSG3241NOUSER) - 1);
    }
    return(err);
}

int
NmapCommandABDelete(void *param)
{
    int err = -1;
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    ContactInfo *pContact;
    unsigned char guid[NMAP_GUID_LENGTH];
    unsigned long id = 0;
    unsigned char *ptr;
    unsigned long indx;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        if ((err = CheckBook(pClient)) == 0) {
            err = 0;
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    if (0 == err) {
        guid[0] = '\0';
        ptr = pClient->buffer + (sizeof(NMAP_ABINFO_COMMAND) - 1);
        if (*ptr != '\0') {
            err = -1;
            ptr++;
            if (*ptr == 'g') {
                ptr++;
                memcpy(guid, ptr, NMAP_GUID_LENGTH);
                guid[NMAP_GUID_LENGTH] = '\0';
                err = 0;
            } else if (isdigit(*ptr)) {
                id = atol(ptr);
                if ((id > 0) && (id <= pBook->control.storeCount)) {
                    err = 0;
                }
            }
        }
        if (err == 0) {
            if (guid[0] != '\0') {
                pContact = NULL;
                for (indx = 0; indx < pIndex->size; indx++) {
                    if (memcmp(
                            pIndex->pList[indx]->guid,
                            guid,
                            NMAP_GUID_LENGTH) == 0) {
                        pContact = pIndex->pList[indx];
                        pContact->flags |= ContactDeleted;
                        err = IndexUpdateFlags(pClient, pContact);
                        if (0 == err) {
                            ConnWrite(
                                    pClient->conn,
                                    MSG1000OK,
                                    sizeof(MSG1000OK) - 1);
                        }
                        break;
                    }
                }
                if (NULL == pContact) {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4220NOMSG,
                            sizeof(MSG4220NOMSG) - 1);
                }
            } else {
                if (id > 0) {
                    pContact = pIndex->pList[id - 1];
                    pContact->flags |= ContactDeleted;
                    err = IndexUpdateFlags(pClient, pContact);
                    if (0 == err) {
                        ConnWrite(
                                pClient->conn,
                                MSG1000OK,
                                sizeof(MSG1000OK) - 1);
                    }
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG3010BADARGC,
                    sizeof(MSG3010BADARGC) - 1);
        }
    }
    return(err);
}

int
NmapCommandABInfo(void *param)
{
    int err = -1;
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    ContactInfo *pContact;
    unsigned char guid[NMAP_GUID_LENGTH];
    unsigned long id = 0;
    unsigned char *ptr;
    unsigned long indx;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        if ((err = CheckBook(pClient)) == 0) {
            err = 0;
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    if (0 == err) {
        guid[0] = '\0';
        ptr = pClient->buffer + (sizeof(NMAP_ABINFO_COMMAND) - 1);
        if (*ptr != '\0') {
            err = -1;
            ptr++;
            if (*ptr == 'g') {
                ptr++;
                memcpy(guid, ptr, NMAP_GUID_LENGTH);
                guid[NMAP_GUID_LENGTH] = '\0';
                err = 0;
            } else if (isdigit(*ptr)) {
                id = atol(ptr);
                if ((id > 0) && (id <= pBook->control.storeCount)) {
                    err = 0;
                }
            }
        }
        if (err == 0) {
            if (guid[0] != '\0') {
                pContact = NULL;
                for (indx = 0; indx < pIndex->size; indx++) {
                    if (memcmp(
                            pIndex->pList[indx]->guid,
                            guid,
                            NMAP_GUID_LENGTH) == 0) {
                        pContact = pIndex->pList[indx];
                        break;
                    }
                }
                if (pContact != NULL) {
                    err = ConnWriteF(
                            pClient->conn,
                            "2001 %ld %ld %ld %s %ld %f\r\n",
                            pContact->id,
                            pContact->size,
                            pContact->flags,
                            pContact->guid,
                            pBook->control.storeId,
                            pContact->modified);
                } else {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4220NOMSG,
                            sizeof(MSG4220NOMSG) - 1);
                }
            } else {
                if (id > 0) {
                    pContact = pIndex->pList[id - 1];
                    if (pContact != NULL) {
                        err = ConnWriteF(
                                pClient->conn,
                                "2001 %ld %ld %ld %s %ld %f\r\n",
                                pContact->id,
                                pContact->size,
                                pContact->flags,
                                pContact->guid,
                                pBook->control.storeId,
                                pContact->modified);
                    } else {
                        err = ConnWrite(
                                pClient->conn,
                                MSG4220NOMSG,
                                sizeof(MSG4220NOMSG) - 1);
                    }
                } else {
                    if (pIndex->next > 0) {
                        ConnWriteF(
                                pClient->conn,
                                "2002 %lu\r\n",
                                pIndex->next);
                        for (indx = 0; indx < pIndex->next; indx++) {
                            pContact = pIndex->pList[indx];
                            if (pContact != NULL) {
                                err = ConnWriteF(
                                        pClient->conn,
                                        "%ld %ld %ld %ld %s %ld %f\r\n",
                                        indx + 1,
                                        pContact->id,
                                        pContact->size,
                                        pContact->flags,
                                        pContact->guid,
                                        pBook->control.storeId,
                                        pContact->modified);
                            }
                        }
                        ConnWrite(
                                pClient->conn,
                                MSG1000OK,
                                sizeof(MSG1000OK) - 1);
                    } else {
                        err = ConnWrite(
                                pClient->conn,
                                MSG4220NOMSG,
                                sizeof(MSG4220NOMSG) - 1);
                    }
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG3010BADARGC,
                    sizeof(MSG3010BADARGC) - 1);
        }
    }
    return(err);
}

int
NmapCommandABList(void *param)
{
    int err = -1;
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    ContactInfo *pContact;
    unsigned char guid[NMAP_GUID_LENGTH];
    unsigned long id = 0;
    unsigned char *ptr;
    unsigned long indx;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        if ((err = CheckBook(pClient)) == 0) {
            err = 0;
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    if ((err == 0) && (pClient->states & NMAP_CLIENT_ABOOK)) {
        guid[0] = '\0';
        ptr = pClient->buffer + (sizeof(NMAP_ABINFO_COMMAND) - 1);
        if (*ptr != '\0') {
            err = -1;
            ptr++;
            if (*ptr == 'g') {
                ptr++;
                memcpy(guid, ptr, NMAP_GUID_LENGTH);
                guid[NMAP_GUID_LENGTH] = '\0';
                err = 0;
            } else if (isdigit(*ptr)) {
                id = atol(ptr);
                if ((id > 0) && (id <= pBook->control.storeCount)) {
                    err = 0;
                }
            }
            if (err == 0) {
                if (guid[0] != '\0') {
                    pContact = NULL;
                    for (indx = 0; indx < pIndex->size; indx++) {
                        if (memcmp(
                                pIndex->pList[indx]->guid,
                                guid,
                                NMAP_GUID_LENGTH) == 0) {
                            pContact = pIndex->pList[indx];
                            break;
                        }
                    }
                    if (pContact != NULL) {
                        err = ConnWriteF(
                                pClient->conn,
                                "2003 %ld\r\n",
                                pContact->size);
                        if (err != -1) {
                            sprintf(
                                    pClient->path,
                                    "%s/%s/%s"NMAP_ADBK_EXT,
                                    pClient->store,
                                    pClient->user,
                                    pBook->name);
                            pBook->hFile = fopen(pClient->path, "rb");
                            if (pBook->hFile != NULL) {
                                err = ConnWriteF(
                                        pClient->conn,
                                        "2023 %lu\r\n",
                                        pContact->size);
                                fseek(
                                        pBook->hFile,
                                        pContact->record + pContact->start,
                                        SEEK_SET);
                                err = ConnWriteFromFile(
                                        pClient->conn,
                                        pBook->hFile,
                                        pContact->size);
                                if (err != -1) {
                                    err = ConnWrite(
                                            pClient->conn,
                                            MSG1000OK,
                                            sizeof(MSG1000OK) - 1);
                                }
                                fclose(pBook->hFile);
                                pBook->hFile = NULL;
                            } else {
                                err = ConnWrite(
                                        pClient->conn,
                                        MSG4225NOADBK,
                                        sizeof(MSG4225NOADBK) - 1);
                            }
                        }
                    } else {
                        err = ConnWrite(
                                pClient->conn,
                                MSG4220NOMSG,
                                sizeof(MSG4220NOMSG) - 1);
                    }
                } else {
                    pContact = pIndex->pList[id - 1];
                    sprintf(
                            pClient->path,
                            "%s/%s/%s"NMAP_ADBK_EXT,
                            pClient->store,
                            pClient->user,
                            pBook->name);
                    pBook->hFile = fopen(pClient->path, "rb");
                    if (pBook->hFile != NULL) {
                        err = ConnWriteF(
                                pClient->conn,
                                "2023 %lu\r\n",
                                pContact->size);
                        fseek(
                                pBook->hFile,
                                pContact->record + pContact->start,
                                SEEK_SET);
                        err = ConnWriteFromFile(
                                pClient->conn,
                                pBook->hFile,
                                pContact->size);
                        if (err != -1) {
                            err = ConnWrite(
                                    pClient->conn,
                                    MSG1000OK,
                                    sizeof(MSG1000OK) - 1);
                        }
                        fclose(pBook->hFile);
                        pBook->hFile = NULL;
                    } else {
                        err = ConnWrite(
                                pClient->conn,
                                MSG4225NOADBK,
                                sizeof(MSG4225NOADBK) - 1);
                    }
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG3010BADARGC,
                    sizeof(MSG3010BADARGC) - 1);
        }
    }
    return(err);
}

int
NmapCommandABPurge(void *param)
{
    int err = -1;
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    ContactInfo *pContact;
    unsigned char guid[NMAP_GUID_LENGTH];
    unsigned long id = 0;
    unsigned char *ptr;
    unsigned long indx;
    unsigned long purgeCount;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        if ((err = CheckBook(pClient)) == 0) {
            err = 0;
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    if (0 == err) {
        guid[0] = '\0';
        ptr = pClient->buffer + (sizeof(NMAP_ABINFO_COMMAND) - 1);
        if (*ptr != '\0') {
            err = -1;
            ptr++;
            if (*ptr == 'g') {
                ptr++;
                memcpy(guid, ptr, NMAP_GUID_LENGTH);
                guid[NMAP_GUID_LENGTH] = '\0';
                err = 0;
            } else if (isdigit(*ptr)) {
                id = atol(ptr);
                if ((id > 0) && (id <= pBook->control.storeCount)) {
                    err = 0;
                }
            }
        }
        if (err == 0) {
            if (guid[0] != '\0') {
                pContact = NULL;
                for (indx = 0; indx < pIndex->size; indx++) {
                    if (memcmp(
                            pIndex->pList[indx]->guid,
                            guid,
                            NMAP_GUID_LENGTH) == 0) {
                        pContact = pIndex->pList[indx];
                        pContact->flags |= ContactPurged;
                        err = IndexUpdateFlags(pClient, pContact);
                        if (0 == err) {
                            err = ConnWriteF(
                                    pClient->conn,
                                    "1000 %s\r\n",
                                    MSG1000PURGED);
                        }
                        break;
                    }
                }
                if (NULL == pContact) {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4220NOMSG,
                            sizeof(MSG4220NOMSG) - 1);
                }
            } else {
                if (id > 0) {
                    pContact = pIndex->pList[id - 1];
                    pContact->flags |= ContactPurged;
                    err = IndexUpdateFlags(pClient, pContact);
                    if (0 == err) {
                        err = ConnWriteF(
                                pClient->conn,
                                "1000 %s\r\n",
                                MSG1000PURGED);
                    }
                } else {
                    err = WriteNLockAquire(pBook->lock, 2000);
                    if (err == NLOCK_AQUIRED) {
                        if ((err = TransBegin(pClient)) == 0) {
                            if ((err = NmapSqlPrepare(
                                            pBook->handle,
                                            INDEX_UPDATE_FLAGS,
                                            &pBook->sql.index.updateFlags)) ==
                                    SQLITE_OK) {
                                for (indx = 0; indx < pIndex->size; indx++) {
                                    pContact = pIndex->pList[indx];
                                    pContact->flags &= ~ContactPurged;
                                    purgeCount++;
                                    sqlite3_bind_int(
                                            pBook->sql.index.
                                            updateFlags,
                                            1,
                                            pContact->flags);
                                    sqlite3_bind_int(
                                            pBook->sql.index.
                                            updateFlags,
                                            2,
                                            pContact->id);
                                    if ((err = Step(
                                            pBook->sql.index.
                                            updateFlags)) == SQLITE_DONE) {
                                        sqlite3_reset(
                                                pBook->sql.index.updateFlags);
                                    } else {
                                        break;
                                    }
                                }
                            }
                        }
                        if (0 == err) {
                            err = TransCommit(pClient);
                        } else {
                            TransAbort(pClient);
                        }
                        if (0 == err) {
                            err = ConnWriteF(
                                    pClient->conn,
                                    "1000 %lu %s\r\n",
                                    purgeCount,
                                    MSG1000PURGED);
                        } else {
                            err = ConnWrite(
                                    pClient->conn,
                                    MSG5005SQLLIBERR,
                                    sizeof(MSG5005SQLLIBERR) - 1);
                        }
                        WriteNLockRelease(pBook->lock);
                    }
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG3010BADARGC,
                    sizeof(MSG3010BADARGC) - 1);
        }
    }
    return(err);
}

int
NmapCommandABRemove(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    unsigned char *ptr;
    int err;

    if (pClient->states & NMAP_CLIENT_USER) {
        if (pClient->states & NMAP_CLIENT_ABOOK) {
            if ((err = CheckBook(pClient)) == 0) {
                err = 0;
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG4225NOADBK,
                    sizeof(MSG4225NOADBK) - 1);
        }
        if (0 == err) {
            ptr = pClient->buffer + (sizeof(NMAP_ABRMOV_COMMAND) - 1);
            if (*ptr != '\0') {
                ptr++;
                if (XplStrCaseCmp(pBook->name, ptr) != 0) {
                    /*
                    ** TODO implement this function
                    **
                    ** acquire an purge lock and delete the files
                    */
                    err = ConnWrite(
                            pClient->conn,
                            MSG1000NOTIMP,
                            sizeof(MSG1000NOTIMP) - 1);
                } else {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4227TGTADBKSELECTED,
                            sizeof(MSG4227TGTADBKSELECTED) - 1);
                }
            }
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG3241NOUSER,
                sizeof(MSG3241NOUSER) - 1);
    }
    return(err);
}

int
NmapCommandABRename(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    int err;

    /*
    ** TODO - ABRename
    **
    ** acquire an purge lock and rename the files
    */
    err = ConnWrite(
            pClient->conn,
            MSG1000NOTIMP,
            sizeof(MSG1000NOTIMP) - 1);
    return(err);
}

int
NmapCommandABSalvage(void *param)
{
    int err = -1;
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    ContactInfo *pContact;
    unsigned char guid[NMAP_GUID_LENGTH];
    unsigned long id = 0;
    unsigned char *ptr;
    unsigned long indx;
    unsigned long delCount = 0;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        if ((err = CheckBook(pClient)) == 0) {
            err = 0;
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    if (0 == err) {
        guid[0] = '\0';
        ptr = pClient->buffer + (sizeof(NMAP_ABINFO_COMMAND) - 1);
        if (*ptr != '\0') {
            err = -1;
            ptr++;
            if (*ptr == 'g') {
                ptr++;
                memcpy(guid, ptr, NMAP_GUID_LENGTH);
                guid[NMAP_GUID_LENGTH] = '\0';
                err = 0;
            } else if (isdigit(*ptr)) {
                id = atol(ptr);
                if ((id > 0) && (id <= pBook->control.storeCount)) {
                    err = 0;
                }
            }
        }
        if (err == 0) {
            if (guid[0] != '\0') {
                pContact = NULL;
                for (indx = 0; indx < pIndex->size; indx++) {
                    if (memcmp(
                            pIndex->pList[indx]->guid,
                            guid,
                            NMAP_GUID_LENGTH) == 0) {
                        pContact = pIndex->pList[indx];
                        pContact->flags &= ~ContactDeleted;
                        err = IndexUpdateFlags(pClient, pContact);
                        if (0 == err) {
                            err = ConnWriteF(
                                    pClient->conn,
                                    "1000 %s\r\n",
                                    MSG1000SALVAGED);
                        }
                        break;
                    }
                }
                if (NULL == pContact) {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4220NOMSG,
                            sizeof(MSG4220NOMSG) - 1);
                }
            } else {
                if (id > 0) {
                    pContact = pIndex->pList[id - 1];
                    pContact->flags &= ~ContactDeleted;
                    err = IndexUpdateFlags(pClient, pContact);
                    if (0 == err) {
                        err = ConnWriteF(
                                pClient->conn,
                                "1000 %s\r\n",
                                MSG1000SALVAGED);
                    }
                } else {
                    err = WriteNLockAquire(pBook->lock, 2000);
                    if (err == NLOCK_AQUIRED) {
                        if ((err = TransBegin(pClient)) == 0) {
                            if ((err = NmapSqlPrepare(
                                            pBook->handle,
                                            INDEX_UPDATE_FLAGS,
                                            &pBook->sql.index.updateFlags)) ==
                                    SQLITE_OK) {
                                for (indx = 0; indx < pIndex->size; indx++) {
                                    pContact = pIndex->pList[indx];
                                    pContact->flags &= ~ContactDeleted;
                                    delCount++;
                                    sqlite3_bind_int(
                                            pBook->sql.index.updateFlags,
                                            1,
                                            pContact->flags);
                                    sqlite3_bind_int(
                                            pBook->sql.index.updateFlags,
                                            2,
                                            pContact->id);
                                    err = Step(pBook->sql.index.updateFlags);
                                    if (err == SQLITE_DONE) {
                                        sqlite3_reset(
                                                pBook->sql.index.updateFlags);
                                    } else {
                                        break;
                                    }
                                }
                            }
                        }
                        if (0 == err) {
                            err = TransCommit(pClient);
                        } else {
                            TransAbort(pClient);
                        }
                        if (0 == err) {
                            err = ConnWriteF(
                                    pClient->conn,
                                    "1000 %lu %s\r\n",
                                    delCount,
                                    MSG1000SALVAGED);
                        } else {
                            err = ConnWrite(
                                    pClient->conn,
                                    MSG5005SQLLIBERR,
                                    sizeof(MSG5005SQLLIBERR) - 1);
                        }
                        WriteNLockRelease(pBook->lock);
                    }
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG3010BADARGC,
                    sizeof(MSG3010BADARGC) - 1);
        }
    }
    return(err);
}

int
NmapCommandABShow(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    int err;

    /*
    ** TODO - ABShow
    */
    err = ConnWrite(
            pClient->conn,
            MSG1000NOTIMP,
            sizeof(MSG1000NOTIMP) - 1);
    return(err);
}

int
NmapCommandABStat(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    int err;

    /*
    ** TODO - ABStat
    */
    err = ConnWrite(
            pClient->conn,
            MSG1000NOTIMP,
            sizeof(MSG1000NOTIMP) - 1);
    return(err);
}

int
NmapCommandABStore(void *param)
{
    int err = -1;
    NMAPClient *pClient = (NMAPClient *) param;
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactInfo *pContact;
    unsigned long readLen = 0;
    unsigned long size = 0;
    unsigned char *ptr;
    unsigned char *pEnd;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        if ((err = CheckBook(pClient)) == 0) {
            err = 0;
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }

    if (0 == err) {
        if (!(NMAP.flags & NMAP_FLAG_DISK_SPACE_LOW)) {
            ptr = pClient->buffer + (sizeof(NMAP_ABSTOR_COMMAND) - 1);
            pEnd = strchr(ptr + 1, ' ');
            if (pEnd == NULL) {
                unsigned char *pCard;

                size = atol(++ptr);
                pCard = MemMalloc(size);
                if (pCard != NULL) {
                    err = ConnWriteF(
                            pClient->conn,
                            "2002 %s\r\n",
                            "Send vCard");
                    if (err != -1) {
                        err = ConnFlush(pClient->conn);
                        if (err != -1) {
                            /*
                            ** TODO Rewrite this section to go back to the
                            ** my original read-to-file code to deal with the
                            ** possibility of large vcards.
                            ** vCard may contain photos and logos which
                            ** could be quite large.
                            **
                            ** Also modify ParseVCard to return to the method
                            ** of parse-from-file.
                            */
                            readLen = ConnReadCount(pClient->conn, pCard, size);
                            if (readLen == size) {
                                pCard[readLen] = '\0';
                                pContact = MemMalloc(sizeof(ContactInfo));
                                if (pContact != NULL) {
                                    memset(pContact, 0, sizeof(ContactInfo));
                                    pContact->size = size;
                                    err = ParseVCard(pClient, pContact, pCard);
                                    if (err == 0) {
                                        err = WriteNLockAquire(
                                                pBook->lock,
                                                2000);
                                        if (err == NLOCK_AQUIRED) {
                                            sprintf(
                                                    pClient->path,
                                                    "%s/%s/%s"NMAP_ADBK_EXT,
                                                    pClient->store,
                                                    pClient->user,
                                                    pBook->name);
                                            pBook->hFile =
                                                    fopen(pClient->path, "ab");
                                            if (pBook->hFile != NULL) {
                                                int offset;

                                                fseek(
                                                       pBook->hFile,
                                                       pBook->control.storeSize,
                                                       SEEK_SET);
                                                pContact->record =
                                                        ftell(pBook->hFile);
                                                GuidAlloc(pContact->guid);
                                                offset = fprintf(
                                                        pBook->hFile,
                                                        "nmap.guid=%s\r\n",
                                                        pContact->guid);
                                                offset += fprintf(
                                                       pBook->hFile,
                                                       "nmap.length=%010lu\r\n",
                                                       pContact->size);
                                                fwrite(
                                                        pCard,
                                                        sizeof(unsigned char),
                                                        pContact->size,
                                                        pBook->hFile);
                                                pContact->start += offset;
                                                pContact->fn += offset;
                                                pContact->n += offset;
                                                pBook->control.storeSize +=
                                                        (pContact->size +
                                                        offset);
                                                pBook->control.storeCount++;
                                                err = ControlUpdate(pClient);
                                                if (err == 0) {
                                                    err = IndexWrite(
                                                            pClient,
                                                            pContact);
                                                    if (err == 0) {
                                                        err = IndexAdd(
                                                                pClient,
                                                                pContact);
                                                        if (0 == err) {
                                                            pContact = NULL;
                                                        }
                                                    }
                                                }
                                                fclose(pBook->hFile);
                                            }
                                            if (err == 0) {
                                                err = ConnWrite(
                                                        pClient->conn,
                                                        MSG1000OK,
                                                        sizeof(MSG1000OK) - 1);
                                            } else {
                                                err = ConnWrite(
                                                        pClient->conn,
                                                        MSG4240NOPERMISSION,
                                                        sizeof
                                                        (MSG4240NOPERMISSION)
                                                        - 1);
                                            }
                                            WriteNLockRelease(pBook->lock);
                                        } else {
                                            err = ConnWrite(
                                                    pClient->conn,
                                                    MSG4120ADBKLOCKED,
                                                    sizeof(MSG4120ADBKLOCKED)
                                                    - 1);
                                        }
                                    } else {
                                        err = ConnWrite(
                                                pClient->conn,
                                                MSG3010BADARGC,
                                                sizeof(MSG3010BADARGC) - 1);
                                    }
                                    if (pContact != NULL) {
                                        MemFree(pContact);
                                    }
                                }
                            } else {
                                err = -1;
                                /*
                                ** TODO INCOMPLETE VCARD error
                                */
                            }
                        }
                    }
                } else {
                    err = ConnWrite(
                            pClient->conn,
                            MSG5001NOMEMORY,
                            sizeof(MSG5001NOMEMORY) - 1);
                }
            } else {
                err = ConnWrite(
                        pClient->conn,
                        MSG3010BADARGC,
                        sizeof(MSG3010BADARGC) - 1);
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG5221SPACELOW,
                    sizeof(MSG5221SPACELOW) - 1);
        }
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    return(err);
}

int
NmapCommandABUpdate(void *param)
{
    NMAPClient *pClient = (NMAPClient *) param;
    int err;

    /*
    ** TODO - write ABUpdate
    */
    err = ConnWrite(
            pClient->conn,
            MSG1000NOTIMP,
            sizeof(MSG1000NOTIMP) - 1);
    return(err);
}

static int
OpenBook(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    if (0 == XplStrCaseCmp(pBook->name, "PERSONAL")) {
        strcpy(pBook->name, "PERSONAL");
        pBook->cb.type = ContactPersonal;
    } else if (0 == XplStrCaseCmp(pBook->name, "SYSTEM")) {
        strcpy(pBook->name, "SYSTEM");
        pBook->cb.type = ContactSystem | ContactReadOnly;
    } else {
        pBook->cb.type = ContactUnknown;
    }
    sprintf(
            pClient->path,
            "%s/%s/%s"NMAP_ADBK_EXT,
            pClient->store,
            pClient->user,
            pBook->name);
    pBook->hFile = fopen(pClient->path, "rb");
    if ((pBook->hFile == NULL) && (pBook->cb.type & ContactPersonal)) {
        err = CreateBook(pClient);
        if (0 == err) {
            pBook->hFile = fopen(pClient->path, "rb");
            if (NULL == pBook->hFile) {
                err = ConnWrite(
                        pClient->conn,
                        MSG4240NOPERMISSION,
                        sizeof(MSG4240NOPERMISSION) - 1);
            }
        }
    }
    if ((0 == err) && (pBook->hFile != NULL)) {
        sprintf(
                pClient->path,
                "%s/%s/%s"NMAP_ADBK_INDEX_EXT,
                pClient->store,
                pClient->user,
                pBook->name);
        if (access(pClient->path, 0) != 0) {
            err = IndexCreate(pClient);
        }
        if (0 == err) {
            err = IndexOpen(pClient);
            if (err == 0) {
                err = IndexRead(pClient);
            }
        }
        fclose(pBook->hFile);
        pBook->hFile = NULL;
    } else {
        /* TODO deal with the Share stuff here */
        err = ConnWrite(
                pClient->conn,
                MSG4224CANTREADADBK,
                sizeof(MSG4224CANTREADADBK) - 1);
    }
    return(err);
}

static int
CreateBook(NMAPClient *pClient)
{
    FILE *hFile;
    int err = -1;

    hFile = fopen(pClient->path, "wb");
    if (NULL != hFile) {
        fclose(hFile);
        err = 0;
    }
    return(err);
}

static int
IndexCreate(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    unlink(pClient->path);
    err = sqlite3_open(pClient->path, &pBook->handle);
    if (SQLITE_OK == err) {
        /*
        ** TODO add code to encase create within a transaction
        */
        err = ControlTableCreate(pClient, &pBook->control);
        if (0 == err) {
            err = IndexTableCreate(pClient);
        }
    }
    return(err);
}

static int
IndexOpen(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactControl *pControl = &pBook->control;
    int err = 0;
    unsigned char abkPath[XPL_MAX_PATH + 1];
    unsigned char dbaPath[XPL_MAX_PATH + 1];
    struct stat sb;

    if (pBook->handle == NULL) {
        sprintf(
                abkPath,
                "%s/%s/%s"NMAP_ADBK_EXT,
                pClient->store,
                pClient->user,
                pBook->name);
        if (stat(abkPath, &sb) == 0) {
            pControl->fileSize = sb.st_size;
            sprintf(
                    dbaPath,
                    "%s/%s/%s"NMAP_ADBK_INDEX_EXT,
                    pClient->store,
                    pClient->user,
                    pBook->name);
            if (pBook->handle == NULL) {
                if (sqlite3_open(dbaPath, &pBook->handle) != SQLITE_OK) {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4240NOPERMISSION,
                            sizeof(MSG4240NOPERMISSION) - 1);
                }
            }
        } else {
            err = ConnWrite(
                    pClient->conn,
                    MSG4240NOPERMISSION,
                    sizeof(MSG4240NOPERMISSION) - 1);
        }
    }
    sprintf(
            pClient->path,
            "%s/%s/%s"NMAP_ADBK_EXT,
            pClient->store,
            pClient->user,
            pBook->name);
    if (stat(pClient->path, &sb) == 0) {
        pControl->fileSize = sb.st_size;
    }
    return(err);
}

static int
IndexRead(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactControl *pControl = &pBook->control;
    ContactControl control;
    int err;

    memset(&control, 0, sizeof(ContactControl));
    if ((err = ControlRead(pClient, &control)) == 0) {
        if (pControl->fileSize != control.storeSize) {
            ControlTableDelete(pClient);
            if ((err = ControlTableCreate(pClient, pControl)) == 0) {
                IndexTableDelete(pClient);
                if ((err = IndexTableCreate(pClient)) == 0) {
                    err = IndexBuild(pClient);
                } else {
                    err = ConnWrite(
                            pClient->conn,
                            MSG4240NOPERMISSION,
                            sizeof(MSG4240NOPERMISSION) - 1);
                }
            } else {
                err = ConnWrite(
                        pClient->conn,
                        MSG4240NOPERMISSION,
                        sizeof(MSG4240NOPERMISSION) - 1);
            }
        } else if ((pControl->id != control.id) ||
                (pControl->version != control.version) ||
                (pControl->storeId != control.storeId) ||
                (pControl->storeSize != control.storeSize) ||
                (pControl->storeCount != control.storeCount) ||
                (pControl->modified != control.modified)) {
            *pControl = control;
            FreeIndex(pClient);
            err = IndexLoad(pClient);
        }
    }
    return(err);
}

void
ContactClose(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;

    /*
    ** TODO Do the actual purge here
    */

    /*
    ** these are from a structure.  Not a good idea to use
    ** free them with a loop, treating the structure as an 
    ** array of statement pointers
    */
    if (pBook->sql.begin) {
        sqlite3_finalize(pBook->sql.begin);
        pBook->sql.begin = NULL;
    }
    if (pBook->sql.commit) {
        sqlite3_finalize(pBook->sql.commit);
        pBook->sql.commit = NULL;
    }
    if (pBook->sql.abort) {
        sqlite3_finalize(pBook->sql.abort);
        pBook->sql.abort = NULL;
    }

    if (pBook->sql.control.read) {
        sqlite3_finalize(pBook->sql.control.read);
        pBook->sql.control.read = NULL;
    }
    if (pBook->sql.control.insertInit) {
        sqlite3_finalize(pBook->sql.control.insertInit);
        pBook->sql.control.insertInit = NULL;
    }
    if (pBook->sql.control.update) {
        sqlite3_finalize(pBook->sql.control.update);
        pBook->sql.control.update = NULL;
    }
    if (pBook->sql.control.drop) {
        sqlite3_finalize(pBook->sql.control.drop);
        pBook->sql.control.drop = NULL;
    }

    if (pBook->sql.index.readAll) {
        sqlite3_finalize(pBook->sql.index.readAll);
        pBook->sql.index.readAll = NULL;
    }
    if (pBook->sql.index.modify) {
        sqlite3_finalize(pBook->sql.index.modify);
        pBook->sql.index.modify = NULL;
    }
    if (pBook->sql.index.insert) {
        sqlite3_finalize(pBook->sql.index.insert);
        pBook->sql.index.insert = NULL;
    }
    if (pBook->sql.index.updateFlags) {
        sqlite3_finalize(pBook->sql.index.updateFlags);
        pBook->sql.index.updateFlags = NULL;
    }
    if (pBook->sql.index.drop) {
        sqlite3_finalize(pBook->sql.index.drop);
        pBook->sql.index.drop = NULL;
    }
    if (pBook->sql.index.tableCreate) {
        sqlite3_finalize(pBook->sql.index.tableCreate);
        pBook->sql.index.tableCreate = NULL;
    }

    if (pBook->handle != NULL) {
        sqlite3_close(pBook->handle);
        pBook->handle = NULL;
    }
}

void
ContactCloseShared(NMAPClient *pClient)
{
    /*
    ** TODO implement the shared stuff
    */
}

static int
ControlRead(NMAPClient *pClient, ContactControl *pControl)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    if ((err = NmapSqlPrepare(
                    pBook->handle,
                    CONTROL_READ,
                    &pBook->sql.control.read)) == SQLITE_OK) {
        err = Step(pBook->sql.control.read);
        if (err == SQLITE_ROW) {
            pControl->id = sqlite3_column_int(pBook->sql.control.read, 0);
            pControl->version =
                    sqlite3_column_int(pBook->sql.control.read, 2);
            if (NMAP_ADBK_VERSION == pControl->version) {
                pControl->flags =
                        sqlite3_column_int(pBook->sql.control.read, 1);
                pControl->storeId =
                        sqlite3_column_int(pBook->sql.control.read, 3);
                pControl->storeSize =
                        sqlite3_column_int(pBook->sql.control.read, 4);
                pControl->storeCount =
                        sqlite3_column_int(pBook->sql.control.read, 5);
                pControl->modified =
                        sqlite3_column_double(pBook->sql.control.read, 6);
                err = SQLITE_OK;
            } else {
                err = SQLITE_FORMAT;
            }
            sqlite3_reset(pBook->sql.control.read);
        }
    }
    return(err);
}

static int
ControlTableDelete(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    if (pClient->states & NMAP_CLIENT_ABOOK) {
        err = 0;
    } else {
        err = ConnWrite(
                pClient->conn,
                MSG4225NOADBK,
                sizeof(MSG4225NOADBK) - 1);
    }
    if (0 == err) {
        if ((err = NmapSqlPrepare(
                        pBook->handle,
                        CONTROL_TABLE_DELETE,
                        &pBook->sql.control.drop)) == SQLITE_OK) {
            if ((err = Step(pBook->sql.control.drop)) == SQLITE_DONE) {
                err = SQLITE_OK;
            }
            sqlite3_reset(pBook->sql.control.drop);
        }
    }
    return(err);
}

static int
ControlTableCreate(NMAPClient *pClient, ContactControl *pControl)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;
    MDBValueStruct *vs;

    if ((err = Execute(pClient, CONTROL_TABLE_CREATE)) == SQLITE_OK) {
        vs = MDBCreateValueStruct(NMAP.handle.directory, NULL);
        if (NULL != vs) {
            pControl->storeId = NMAP.universalCounter++;
            sprintf(pClient->line, "%lu", NMAP.universalCounter);
            MDBAddValue(pClient->line, vs);
            MDBWrite(MSGSRV_AGENT_NMAP, MSGSRV_A_UID, vs);
            MDBDestroyValueStruct(vs);
            if ((err = NmapSqlPrepare(
                            pBook->handle,
                            CONTROL_TABLE_INSERT_INIT,
                            &pBook->sql.control.insertInit)) == SQLITE_OK) {
                pControl->version = NMAP_ADBK_VERSION;
                sqlite3_bind_int(
                        pBook->sql.control.insertInit,
                        1,
                        pControl->version);  /* version */
                sqlite3_bind_int(
                        pBook->sql.control.insertInit,
                        2,
                        pControl->storeId);  /* storeId */
                if ((err = Step(pBook->sql.control.insertInit)) ==
                        SQLITE_DONE) {
                    pControl->id = (unsigned long) sqlite3_last_insert_rowid(
                            pBook->handle);
                    /*
                    ** leave the modified value incorrect. This will cause and
                    ** index read
                    */
                    pControl->modified = 0;
                    err = SQLITE_OK;
                }
                sqlite3_reset(pBook->sql.control.insertInit);
            }
        }
    }
    return(err);
}

static int
ControlUpdate(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactControl *pControl = &pBook->control;
    int err;

    if ((err = NmapSqlPrepare(
                    pBook->handle,
                    CONTROL_UPDATE,
                    &pBook->sql.control.update)) == SQLITE_OK) {
        sqlite3_bind_int(pBook->sql.control.update, 1, pControl->flags);
        sqlite3_bind_int(pBook->sql.control.update, 2, pControl->storeId);
        sqlite3_bind_int(pBook->sql.control.update, 3, pControl->storeSize);
        sqlite3_bind_int(pBook->sql.control.update, 4, pControl->storeCount);
        sqlite3_bind_int(pBook->sql.control.update, 5, pControl->id);
        err = Step(pBook->sql.control.update);
        if (err == SQLITE_DONE) {
            err = 0;
        }
        sqlite3_reset(pBook->sql.control.update);
    }
    return(err);
}

static int
IndexLoad(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactInfo *pContact;
    int err;
    BOOL bDone = FALSE;
    const unsigned char *text;
    unsigned long count = 0;

    if ((err = NmapSqlPrepare(
                    pBook->handle,
                    INDEX_READ_ALL,
                    &pBook->sql.index.readAll)) == SQLITE_OK) {
        while (!bDone) {
            if ((err = Step(pBook->sql.index.readAll)) == SQLITE_ROW) {
                pContact = (ContactInfo *) MemMalloc(sizeof(ContactInfo));
                if (pContact != NULL) {
                    pContact->id =
                            sqlite3_column_int(pBook->sql.index.readAll, 0);
                    pContact->flags =
                            sqlite3_column_int(pBook->sql.index.readAll, 1);
                    text = sqlite3_column_text(pBook->sql.index.readAll, 2);
                    memcpy(pContact->guid, text, NMAP_GUID_LENGTH);
                    pContact->guid[NMAP_GUID_LENGTH] = '\0';
                    pContact->record =
                            sqlite3_column_int(pBook->sql.index.readAll, 3);
                    pContact->start =
                            sqlite3_column_int(pBook->sql.index.readAll, 4);
                    pContact->fn =
                            sqlite3_column_int(pBook->sql.index.readAll, 5);
                    pContact->n =
                            sqlite3_column_int(pBook->sql.index.readAll, 6);
                    pContact->size =
                            sqlite3_column_int(pBook->sql.index.readAll, 7);
                    pContact->indexed =
                            sqlite3_column_double(pBook->sql.index.readAll, 8);
                    pContact->modified =
                            sqlite3_column_double(pBook->sql.index.readAll, 9);
                    /*
                    ** do we show deleted, not purged?
                    */
                    if (!(pContact->flags & ContactDeleted)) {
                        if ((err = IndexAdd(pClient, pContact)) == 0) {
                            count++;
                        } else {
                            MemFree(pContact);
                            bDone = TRUE;
                        }
                    }
                } else {
                    err = ConnWrite(
                            pClient->conn,
                            MSG5001NOMEMORY,
                            sizeof(MSG5001NOMEMORY) - 1);
                    bDone = TRUE;
                }
            } else if (SQLITE_DONE == err) {
                err = SQLITE_OK;
                bDone = TRUE;
            } else {
                bDone = TRUE;
            }
        }
        sqlite3_reset(pBook->sql.index.readAll);
    }
    if (0 == err) {
        if (count != pBook->control.storeCount) {
            pBook->control.storeCount = count;
            ControlUpdate(pClient);
        }
    }
    return(err);
}

static int
IndexTableDelete(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    if ((err = NmapSqlPrepare(
                    pBook->handle,
                    INDEX_TABLE_DELETE,
                    &pBook->sql.index.drop)) == SQLITE_OK) {
        if ((err = Step(pBook->sql.index.drop)) == SQLITE_DONE) {
            err = SQLITE_OK;
        }
        sqlite3_reset(pBook->sql.index.drop);
    }
    return(err);
}

static int
IndexTableCreate(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    if ((err = NmapSqlPrepare(
                    pBook->handle,
                    INDEX_TABLE_CREATE,
                    &pBook->sql.index.tableCreate)) == SQLITE_OK) {
        if ((err = Step(pBook->sql.index.tableCreate)) == SQLITE_DONE) {
            err = SQLITE_OK;
        }
        sqlite3_reset(pBook->sql.index.tableCreate);
    }
    return(err);
}

static int
IndexBuild(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactInfo *pContact;
    int err = 0;
    unsigned char buf[2048 + 1];
    unsigned char *ptr;
    unsigned char *pEnd;
    unsigned long offset = 0;
    BOOL bMalloc = FALSE;
    unsigned char *pCard;

    fseek(pBook->hFile, 0, SEEK_SET);

    err = TransBegin(pClient);
    while (0 == err) {
        offset = ftell(pBook->hFile);
        ptr = fgets(buf, 2048, pBook->hFile);
        if (ptr != NULL) {
            if (('n' == *ptr) && (memcmp(ptr, "nmap.guid=", 10) == 0)) {
                pContact = (ContactInfo *) MemMalloc(sizeof(ContactInfo));
                if (NULL != pContact) {
                    pContact->start = offset;
                    ptr += 5;
                    if ((pEnd = strchr(ptr, '\r')) != NULL) {
                        *pEnd = '\0';
                        memcpy(pContact->guid, ptr, NMAP_GUID_LENGTH);
                        pContact->guid[NMAP_GUID_LENGTH] = '\0';
                    } else {
                        err = -1;      /* format */
                    }
                    if (0 == err) {
                        if ((ptr = fgets(buf, 2048, pBook->hFile)) != NULL) {
                            if (('n' == *ptr) &&
                                    (memcmp(ptr, "nmap.length=", 12) == 0)) {
                                ptr += 5;
                                if ((pEnd = strchr(ptr, '\r')) != NULL) {
                                    *pEnd = '\0';
                                    pContact->size = atol(ptr);
                                } else {
                                    err = -1;   /* format */
                                }
                            } else {
                                err = -1;   /* file format */
                            }
                        } else {
                            err = -3;  /* file length */
                        }
                    }
                    if (0 == err) {
                        if (pContact->size > 2048) {
                            bMalloc = TRUE;
                            pCard = MemMalloc(pContact->size + 1);
                        } else {
                            bMalloc = FALSE;
                            pCard = buf;
                        }
                        if (pCard != NULL) {
                            if ((ptr = fgets(pCard,
                                            pContact->size,
                                            pBook->hFile)) != NULL) {
                                err = ParseVCard(pClient, pContact, pCard);
                                if (err == 0) {
                                    if (err == 0) {
                                        err = IndexWrite(pClient, pContact);
                                        if (err == 0) {
                                            err = IndexAdd(pClient, pContact);
                                            if (0 == err) {
                                                pContact = NULL;
                                            }
                                        }
                                    }
                                }
                            } else {
                                err = -1;
                            }
                            if (bMalloc == TRUE) {
                                MemFree(pCard);
                                bMalloc = FALSE;
                            }
                        } else {
                            err = ConnWrite(
                                    pClient->conn,
                                    MSG5001NOMEMORY,
                                    sizeof(MSG5001NOMEMORY) - 1);
                        }
                    }
                    if (pContact != NULL) {
                        MemFree(pContact);
                    }
                } else {
                    err = -2;          /* memory */
                }
            } else {
                err = -1;              /* file format */
            }
        } else {
            if (feof(pBook->hFile)) {
                break;
            } else {
                err = -1;              /* file format */
            }
        }
    }
    if (err == 0) {
        err = ControlUpdate(pClient);
    }
    if (err == 0) {
        err = TransCommit(pClient);
    } else {
        TransAbort(pClient);
    }
    return(err);
}

static int
IndexAdd(NMAPClient *pClient, ContactInfo *pContact)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    ContactInfo **ppList;
    int err = 0;

    if (pIndex->pList == NULL) {
        pIndex->pList = (ContactInfo **) MemMalloc(16 * sizeof(ContactInfo *));
        if (pIndex->pList != NULL) {
            pIndex->size = 16;
            pIndex->next = 0;
        } else {
            err = -1;                  /* memory */
        }
    } else {
        if (pIndex->next == pIndex->size) {
            ppList = (ContactInfo **) MemRealloc(pIndex->pList,
                    pIndex->size + 16);
            if (ppList != NULL) {
                pIndex->pList = ppList;
                pIndex->size += 16;
            } else {
                err = -1;
            }
        }
    }
    if (0 == err) {
        pIndex->pList[pIndex->next++] = pContact;
    }
    return(err);
}

static int
IndexWrite(NMAPClient *pClient, ContactInfo *pContact)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err = 0;

    if ((err = NmapSqlPrepare(
                    pBook->handle,
                    INDEX_INSERT,
                    &pBook->sql.index.insert)) == SQLITE_OK) {
        sqlite3_bind_int(pBook->sql.index.insert, 1, pContact->flags);
        sqlite3_bind_text(
                pBook->sql.index.insert,
                2,
                pContact->guid,
                NMAP_GUID_LENGTH,
                SQLITE_STATIC);
        sqlite3_bind_int(pBook->sql.index.insert, 3, pContact->record);
        sqlite3_bind_int(pBook->sql.index.insert, 4, pContact->start);
        sqlite3_bind_int(pBook->sql.index.insert, 5, pContact->fn);
        sqlite3_bind_int(pBook->sql.index.insert, 6, pContact->n);
        sqlite3_bind_int(pBook->sql.index.insert, 7, pContact->size);
        sqlite3_bind_double(pBook->sql.index.insert, 8, pContact->indexed);
        err = TransBegin(pClient);
        if (SQLITE_OK == err) {
            if ((err = Step(pBook->sql.index.insert)) == SQLITE_DONE) {
                if ((err = TransCommit(pClient)) == SQLITE_OK) {
                    pContact->id = (unsigned long) sqlite3_last_insert_rowid(
                            pBook->handle);
                    if ((err = NmapSqlPrepare(
                            pBook->handle,
                            "SELECT julianday(modified) FROM contact WHERE id = ?",
                            &pBook->sql.index.modify)) == SQLITE_OK) {
                        sqlite3_bind_int(
                                pBook->sql.index.modify,
                                1,
                                pContact->id);
                        err = Step(pBook->sql.index.modify);
                        if (err == SQLITE_ROW) {
                            pContact->modified = sqlite3_column_double(
                                    pBook->sql.index.
                                    modify,
                                    0);
                            err = 0;
                        }
                        sqlite3_reset(pBook->sql.index.modify);
                    }
                }
            } else {
                TransAbort(pClient);
            }
        }
        sqlite3_reset(pBook->sql.index.insert);
    }
    return(err);
}

static int
TransBegin(NMAPClient *pClient)
{
    int err = 0;
    int indx;

    err = NmapSqlPrepare(
            pClient->addressBook.handle,
            NMAP_SQLTRAN_BEGIN,
            &pClient->addressBook.sql.begin);
    if (err == SQLITE_OK) {
        for (indx = 0; indx < 30; indx++) {
            err = sqlite3_step(pClient->addressBook.sql.begin);
            if (SQLITE_DONE == err) {
                sqlite3_reset(pClient->addressBook.sql.begin);
                err = SQLITE_OK;
                break;
            } else if (SQLITE_BUSY == err) {
                XplDelay(300);
            } else {
                break;
            }
        }
    }
    return(err);
}

static int
TransCommit(NMAPClient *pClient)
{
    int err = 0;
    int indx;

    err = NmapSqlPrepare(
            pClient->addressBook.handle,
            NMAP_SQLTRAN_END,
            &pClient->addressBook.sql.commit);
    if (err == SQLITE_OK) {
        for (indx = 0; indx < 30; indx++) {
            err = sqlite3_step(pClient->addressBook.sql.commit);
            if (SQLITE_DONE == err) {
                sqlite3_reset(pClient->addressBook.sql.commit);
                err = SQLITE_OK;
                break;
            } else if (SQLITE_BUSY == err) {
                XplDelay(300);
            } else {
                break;
            }
        }
    }
    return(err);
}

static int
TransAbort(NMAPClient *pClient)
{
    int err = 0;
    int indx;

    err = NmapSqlPrepare(
            pClient->addressBook.handle,
            NMAP_SQLTRAN_ROLLBACK,
            &pClient->addressBook.sql.abort);
    if (err == SQLITE_OK) {
        for (indx = 0; indx < 30; indx++) {
            err = sqlite3_step(pClient->addressBook.sql.abort);
            if (SQLITE_DONE == err) {
                sqlite3_reset(pClient->addressBook.sql.abort);
                err = SQLITE_OK;
                break;
            } else if (SQLITE_BUSY == err) {
                XplDelay(300);
            } else {
                break;
            }
        }
    }
    return(err);
}

static int
Execute(NMAPClient *pClient, unsigned char *ptr)
{
    int err;
    int indx;

    for (indx = 0; indx < 30; indx++) {
        if ((err = sqlite3_exec(pClient->addressBook.handle,
                        (char *) ptr,
                        NULL,
                        NULL,
                        NULL)) == SQLITE_OK) {
            break;
        } else if (err == SQLITE_BUSY) {
            XplDelay(300);
        } else {
            break;
        }
    }
    return(err);
}

static int
Step(sqlite3_stmt *stmt)
{
    int err;
    int indx;
    BOOL bDone = FALSE;

    for (indx = 0; ((indx < 30) && (!bDone)); indx++) {
        err = sqlite3_step(stmt);
        switch (err) {
            case SQLITE_ROW:
            case SQLITE_DONE:
            default:
                bDone = TRUE;
                break;
            case SQLITE_BUSY:
                XplDelay(300);
                break;
        }
    }
    return(err);
}

static void
FreeIndex(NMAPClient *pClient)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    ContactIndex *pIndex = &pBook->index;
    unsigned long indx;

    if (NULL != pIndex->pList) {
        for (indx = 0; indx < pIndex->next; indx++) {
            MemFree(pIndex->pList[indx]);
        }
        MemFree(pIndex->pList);
        pIndex->pList = NULL;
        pIndex->size = 0;
        pIndex->next = 0;
    }
}

static int
CheckBook(NMAPClient *pClient)
{
    int err = 0;

    if (!(pClient->share.flags & NMAP_SHARE_ADDRESS_BOOK)) {
        err = IndexOpen(pClient);
        if (err == 0) {
            err = IndexRead(pClient);
            if (0 == err) {
                if (pClient->flags & NMAP_OOBMESSAGES) {
                    err = ConnWriteF(
                            pClient->conn,
                            "6000 %d New Contacts\r\n",
                            EVENT_ADBK_NEW_CONTACTS);
                }
            }
        }
    } else if (!IsAvailableShare(pClient->user,
                pClient->addressBook.name,
                NMAP_SHARE_ADDRESS_BOOK,
                NULL,
                NULL)) {
        BOOL bChanged = FALSE;

        if (SyncSharedAddressBook(pClient, &bChanged) &&
                (bChanged) && (pClient->flags & NMAP_OOBMESSAGES)) {
            err = ConnWriteF(
                    pClient->conn,
                    "6000 %d New Contacts\r\n",
                    EVENT_ADBK_NEW_CONTACTS);
        }
    } else {
        pClient->share.flags |= NMAP_SHARE_ADDRESS_BOOK_DENIED;
    }
    return(err);
}

static int
ParseVCard(NMAPClient *pClient, ContactInfo *pContact, unsigned char *pCard)
{
    ContactParse status = ContactVersion | ContactFN | ContactN;
    unsigned char *ptr = pCard;
    BOOL bInCard = FALSE;
    BOOL bDone = FALSE;

    while ((ptr != NULL) && (!bDone)) {
        switch (*ptr) {
            case 'B':
                if (memcmp(ptr, "BEGIN:VCARD", 11) == 0) {
                    bInCard = TRUE;
                }
                break;
            case 'F':
                if ((bInCard == TRUE) && (memcmp(ptr, "FN:", 3) == 0)) {
                    pContact->fn = ptr - pCard;
                    status &= ~ContactFN;
                }
                break;
            case 'N':
                if ((bInCard == TRUE) && (memcmp(ptr, "N:", 2) == 0)) {
                    pContact->n = ptr - pCard;
                    status &= ~ContactN;
                }
                break;
            case 'V':
                if ((bInCard == TRUE) && (memcmp(ptr, "VERSION:", 8) == 0)) {
                    ptr = strchr(ptr, ':') + 1;
                    if (memcmp(ptr, "3.0", 3) == 0) {
                        status &= ~ContactVersion;
                    }
                }
                break;
            case 'E':
                if ((bInCard == TRUE) && (memcmp(ptr, "END:VCARD", 9) == 0)) {
                    bDone = TRUE;
                    break;
                }
                break;
        }
        ptr = strchr(ptr, '\n');
        if (ptr != NULL) {
            ptr++;
        }
    }
    return((int) status);
}

static int
IndexUpdateFlags(NMAPClient *pClient, ContactInfo *pContact)
{
    NmapAddressBook *pBook = &pClient->addressBook;
    int err;

    err = WriteNLockAquire(pBook->lock, 2000);
    if (err == NLOCK_AQUIRED) {
        if ((err = TransBegin(pClient)) == SQLITE_OK) {
            if ((err = NmapSqlPrepare(
                            pBook->handle,
                            INDEX_UPDATE_FLAGS,
                            &pBook->sql.index.updateFlags)) == SQLITE_OK) {
                sqlite3_bind_int(
                        pBook->sql.index.updateFlags,
                        1,
                        pContact->flags);
                sqlite3_bind_int(
                        pBook->sql.index.updateFlags,
                        2,
                        pContact->id);
                if ((err = Step(pBook->sql.index.updateFlags)) == SQLITE_DONE) {
                    err = TransCommit(pClient);
                } else {
                    TransAbort(pClient);
                }
                sqlite3_reset(pBook->sql.index.updateFlags);
            }
        }
        WriteNLockRelease(pBook->lock);
    }
    return(err);
}

/*
** vim:ts=4:sw=4:sts=4:et
*/
