/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2005 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include <config.h>
#include <xpl.h>
#include <logger.h>
#include <mdb.h>
#include <nmap.h>
#include <nmlib.h>

#include "nmapdp.h"

/*
** dstoreauth.c
**
** Contains code to evaluate authorization to access
** store documents
*/


typedef struct _NMAPPrincipal {
    NMAPPrincipalType type;
    union {
        char *user;
        char *email;
    };
} NMAPPrincipal;


typedef struct _NMAPACE {
    NMAPPrincipal principal;
    unsigned int privileges;   /* granted privileges */
} NMAPACE;



/* 
   <user-principal>         ::= 'user:' <username>
   <email-principal>        ::= 'email:' <email-address>
   <principal>              ::= <user-principal> | <email-principal> | 'all' | 'auth'
   <operation>              ::= 'grant' | 'deny'
   <access-control-element> ::= <operation> ' ' <principal> ' ' <privilege> ';'    
   <access-control-list>    ::= <access-control-element> [<access-control-list>]
*/

/* returns:
   -0 success
   - -1 error
   
*/

static int
parseACE (char **ptr, NMAPACE *ace)
{
    char *p = *ptr;
    char *q;

#define GetNextToken(end,str,delim)          \
        if (!(end = strchr(str, delim))) {   \
            return -1;                       \
        } 

    /* operation */
    GetNextToken(q, p, ' ');
    if (!strncmp("grant", p, 5)) {
        
    } else if (!strncmp("deny", p, 4)) {
        return -1; /* unsupported */
    } else {
        return -1;
    }
    p = ++q;
    
    /* principal */
    GetNextToken(q, p, ' ');
    *q = 0;
    if (!strcmp(p, "all")) {
        ace->principal.type = NMAP_PRINCIPAL_ALL;
    } else if (!strcmp(p, "auth")) {
        ace->principal.type = NMAP_PRINCIPAL_AUTHENTICATED;
    } else if (!strncmp(p, "user:", 5)) {
        ace->principal.type = NMAP_PRINCIPAL_USER;
        ace->principal.user = p + 5;
    } else if (!strncmp(p, "email:", 6)) {
        ace->principal.type = NMAP_PRINCIPAL_EMAIL;
        ace->principal.user =  p + 6;
    } else {
        return -1;
    }
    p = ++q;
                
    /* privilege */
    GetNextToken(q, p, ';');
    *q = 0;
    if (!strcmp(p, "all")) {
        ace->privileges = NMAP_PRIV_ALL;
    } else if (!strcmp(p, "read")) {
        ace->privileges = NMAP_PRIV_READ;
    } else if (!strcmp(p, "write")) { /* write -> write-content */
        ace->privileges = NMAP_PRIV_WRITE_CONTENT;
    } else if (!strcmp(p, "write-props")) {
        ace->privileges = NMAP_PRIV_WRITE_PROPS;
    } else if (!strcmp(p, "write-content")) {
        ace->privileges = NMAP_PRIV_WRITE_CONTENT;
    } else if (!strcmp(p, "unlock")) {
        ace->privileges = NMAP_PRIV_UNLOCK;
    } else if (!strcmp(p, "read-acl")) {
        ace->privileges = NMAP_PRIV_READ_ACL;
    } else if (!strcmp(p, "read-own-acl")) {
        ace->privileges = NMAP_PRIV_READ_OWN_ACL;
    } else if (!strcmp(p, "write-acl")) {
        ace->privileges = NMAP_PRIV_WRITE_ACL;
    } else if (!strcmp(p, "bind")) {
        ace->privileges = NMAP_PRIV_BIND;
    } else if (!strcmp(p, "unbind")) {
        ace->privileges = NMAP_PRIV_UNBIND;
    } else {
        return -1;
    }
    
    *ptr = ++q;
    return 0;
}


/* returns: 
   - 0 success 
   - non-zero: error occurred, response sent to client
*/

static int
computePrivileges (NMAPClient *client, char *guid,
                   unsigned int *privileges)
{
    int ccode;
    const char *prop;
    const char *val;
    char buf[CONN_BUFSIZE + 1];
    char *ptr;
    NMAPACE ace;

    if (SQLITE_OK != NmapSqlPrepare(client->dstore.handle, NMAP_DPGET_SELECT, 
                                    &client->dstore.sql.sel.apinfo)
        || SQLITE_OK != sqlite3_bind_text(client->dstore.sql.sel.apinfo, 1, 
                                          guid, NMAP_GUID_LENGTH, SQLITE_STATIC))
    {
        goto abort_sqlite;
    }

    while (SQLITE_ROW == (ccode = sqlite3_step(client->dstore.sql.sel.apinfo))) {
        prop = sqlite3_column_text(client->dstore.sql.sel.apinfo, 0);
        val = sqlite3_column_text(client->dstore.sql.sel.apinfo, 1);
        if (!val || !prop) {
            goto abort_sqlite;
        }
        if (strcmp(prop, "nmap.access-control")) {
            continue;
        }

        strncpy(buf, val, CONN_BUFSIZE);
        buf[CONN_BUFSIZE] = 0;
        ptr = buf;

        do {
            if (parseACE(&ptr, &ace)) {
                sqlite3_reset(client->dstore.sql.sel.apinfo);
                return ConnWrite(client->conn, MSG5006CORRUPTACL, 
                                 sizeof(MSG5006CORRUPTACL) - 1);
            }

            XplConsolePrintf("ACE: grant %d %s %x\r\n", ace.principal.type, ace.principal.user, ace.privileges);
            
            switch (ace.principal.type) {
            case NMAP_PRINCIPAL_ALL:
                break;
            case NMAP_PRINCIPAL_AUTHENTICATED:
                if (!(client->states & NMAP_CLIENT_AUTHENTICATED)) {
                    continue;
                }
                break;
            case NMAP_PRINCIPAL_USER:
                if (NMAP_PRINCIPAL_USER != client->principalType || 
                    strcmp(ace.principal.user, client->principal)) 
                {
                    continue;
                }
                break;
            case NMAP_PRINCIPAL_EMAIL: /* unimplemented */
                if (NMAP_PRINCIPAL_EMAIL != client->principalType || 
                    strcmp(ace.principal.user, client->principal)) 
                {
                    continue;
                }
                break;
            }
            
            *privileges |= ace.privileges;
        } while (*ptr);
    }

    if (SQLITE_DONE != ccode) {
        goto abort_sqlite;
    }
    sqlite3_reset(client->dstore.sql.sel.apinfo);

    return 0;

abort_sqlite:
    XplConsolePrintf ("Sqlite abort: %s\r\n", sqlite3_errmsg(client->dstore.handle));
    return ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
}


/** Checks that the client is authorized to perform the requested 
    operation on the document at guid.

    @param client the connected NMAP client.
    
    @param guid 

    @param silent

    @return 
    - zero if authorized
    - non-zero upon error or if not authorized (mesg printed to client)
    - -4240 if not authorized and mesg not printed to client
*/

int
NmapCheckAuthorization (NMAPClient *client, char *guid, unsigned int neededPrivileges,
                        int silent)
{
    unsigned int privileges = 0;
    int ccode;
    int type;
    char rbuf[XPL_MAX_PATH + 1];
    char *resource = NULL;
    char curguid[NMAP_GUID_LENGTH+1];
    
    if (NMAP_PRINCIPAL_USER == client->principalType && 
        (!client->principal[0] || !strcmp(client->principal, client->user)))
    {
        /* user always has full permissions to her own objects. */
        return 0;
    }

    strncpy(curguid, guid, sizeof(curguid));
    curguid[sizeof(curguid)-1] = 0;

    while (1) {
        ccode = computePrivileges (client, curguid, &privileges);
        if (neededPrivileges == (privileges & neededPrivileges)) {
            return 0;
        }

        XplConsolePrintf("have privs: %x need: %x; checking parent\r\n", 
                         privileges, neededPrivileges);
        
        /* don't have privs, check the parent resource */
        if (!resource) {
            XplConsolePrintf("finding resource for guid %s\r\n", curguid);

            if (SQLITE_OK != NmapSqlPrepare(client->dstore.handle, 
                                            NMAP_DINFO_BY_GUID, 
                                            &client->dstore.sql.sel.arguid)
                || SQLITE_OK != sqlite3_bind_text(client->dstore.sql.sel.arguid, 
                                                  1, curguid, NMAP_GUID_LENGTH, 
                                                  SQLITE_STATIC)
                || SQLITE_ROW != sqlite3_step(client->dstore.sql.sel.arguid))
            {
                XplConsolePrintf("sql error: %s\r\n", 
                                 sqlite3_errmsg(client->dstore.handle));
                return ConnWrite(client->conn, MSG5005SQLLIBERR, 
                                 sizeof(MSG5005SQLLIBERR) - 1);
            }

            strncpy(rbuf, sqlite3_column_text(client->dstore.sql.sel.arguid, 0), 
                    sizeof(rbuf));
            rbuf[sizeof(rbuf) - 1] = 0;
            resource = rbuf;
            type = sqlite3_column_int(client->dstore.sql.sel.arguid, 1);

            sqlite3_reset(client->dstore.sql.sel.arguid);
        }

        XplConsolePrintf("orig type: %x, resource: %s\r\n", type, resource);

        if (NMAP_IS_FOLDER(type)) {
            char *ptr = strrchr(resource, '/');
            if (!ptr) {
                if (silent) {
                    return -4240;
                } else {
                    return ConnWrite(client->conn, MSG4240NOPERMISSION, 
                                     sizeof(MSG4240NOPERMISSION) - 1);
                }
            }
            *ptr = 0;
        } else {
            type |= NMAP_DOCTYPE_FOLDER;
        }

        XplConsolePrintf("new type: %x, resource: %s\r\n", type, resource);

        ccode = FindResourceGuid(client, resource, type, curguid);
        if (ccode) {
            if (-4224 == ccode && !silent) {
                return ConnWrite(client->conn, MSG4240NOPERMISSION, 
                                 sizeof(MSG4240NOPERMISSION) - 1);
                /*
                return ConnWrite(client->conn, MSG4224NOMBOX, 
                                 sizeof(MSG4224NOMBOX) - 1);
                */
            }
            return ccode;
        }
    }
}


/** Finds the guid associated with the specified resource and type.

    @param client the connected NMAP client.
    
    @param resource
    
    @param type

    @param outGuid ptr to a buffer that must be at least NMAP_GUID_LENGTH + 1
    
    @param silent

    @return 
    - zero on success.
    - non-zero upon error (mesg printed to client)
    - -4224 if couldn't find resource (and silent is set)
*/

int
FindResourceGuid (NMAPClient *client, 
                  char *resource, NmapDocumentTypes type, char *outGuid)
{
    XplConsolePrintf ("looking up guid for %s (type %d)\r\n", resource, type);
    
    if (SQLITE_OK != NmapSqlPrepare(client->dstore.handle, 
                                    NMAP_DLIST_FOLDER, 
                                    &client->dstore.sql.sel.aldir)
        || SQLITE_OK != sqlite3_bind_text(client->dstore.sql.sel.aldir, 
                                          1, resource, strlen(resource),
                                          SQLITE_STATIC)
        || SQLITE_OK != sqlite3_bind_int(client->dstore.sql.sel.aldir, 
                                         2, type))
    {
        XplConsolePrintf ("sql error %s\r\n",
                          sqlite3_errmsg(client->dstore.handle));
        return ConnWrite(client->conn, MSG5005SQLLIBERR, 
                         sizeof(MSG5005SQLLIBERR) - 1);
    }
       
    switch (sqlite3_step(client->dstore.sql.sel.aldir)) {
    case SQLITE_DONE:
    {   
        int result;
        char * ptr = strrchr(resource, '/');

        sqlite3_reset(client->dstore.sql.sel.aldir);
        if (!ptr) {
            return -4224;
        }
        *ptr = 0;
        result = FindResourceGuid(client, resource, type, outGuid);
        *ptr = '/';
        return result;
    }
    case SQLITE_ROW:
        strncpy(outGuid, sqlite3_column_text(client->dstore.sql.sel.aldir, 0),
                NMAP_GUID_LENGTH);
        outGuid[NMAP_GUID_LENGTH] = 0;
        sqlite3_reset(client->dstore.sql.sel.aldir);
        return 0;
    default:
        XplConsolePrintf ("sql error %s\r\n",
                          sqlite3_errmsg(client->dstore.handle));
        return ConnWrite(client->conn, MSG5005SQLLIBERR, 
                         sizeof(MSG5005SQLLIBERR) - 1);
    }
}
