/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include <config.h>
#include <xpl.h>
#include <logger.h>

#include "nmapdp.h"

#if defined(WIN32)
#include <rpcdce.h>

void 
GuidReset(void)
{
    int i;
    unsigned long length;
    unsigned char salt[32];
    unsigned char name[MAX_COMPUTERNAME_LENGTH + 1];
    unsigned char digest[SHA_DIGEST_LENGTH];
    unsigned char *uid = NULL;
    unsigned char *ptr;
    FILETIME date;
    SHA_CTX context;
    UUID uuid;

    SHA1_Init(&context);

    GetSystemTimeAsFileTime(&date);
    sprintf(salt, "%010lu%010lu", date.dwHighDateTime, date.dwLowDateTime);
    SHA1_Update(&context, salt, strlen(salt));

    length = sizeof(name) - 1;
    if (GetComputerName(name, &length)) {
        SHA1_Update(&context, name, strlen(name));
    } else {
        RAND_bytes(name, sizeof(name) - 1);
        SHA1_Update(&context, name, sizeof(name) - 1);
    }

    if ((UuidCreate(&uuid) == RPC_S_OK) && (UuidToString(&uuid, &uid) == RPC_S_OK)) {
        SHA1_Update(&context, uid, strlen(uid));
    } else {
        RAND_bytes(name, sizeof(name) - 1);
        SHA1_Update(&context, name, sizeof(name) - 1);
    }

    if (uid) {
        RpcStringFree(&uid);
    }

    SHA1_Final(digest, &context);

    for (i = 0, ptr = NMAP.guid.next; i < SHA_DIGEST_LENGTH; i++, ptr += 2) {
        sprintf(ptr, "%02X", digest[i]);
    }

    memset(NMAP.guid.next + NMAP_GUID_PREFIX_LENGTH, '0', NMAP_GUID_SUFFIX_LENGTH);

    return;
}
#elif defined(NETWARE) || defined(LIBC)
#error GuidReset not implemented on this platform.

void 
GuidReset(void)
{
    return;
}
#elif defined(LINUX)

#if HAVE_SYS_SYSINFO_H
#include <sys/sysinfo.h>
#endif

void 
GuidReset(void)
{
    int i;
    char name[256 + 1];
    unsigned char salt[32];
    unsigned char digest[SHA_DIGEST_LENGTH];
    unsigned char *ptr;
#if HAVE_SYS_SYSINFO_H
    struct sysinfo si;
#endif
    struct timeval tv;
    SHA_CTX context;

    SHA1_Init(&context);

    gettimeofday(&tv, NULL);
    sprintf(salt, "%010lu%010lu", tv.tv_sec, tv.tv_usec);
    SHA1_Update(&context, salt, strlen(salt));

    if (!gethostname(name, sizeof(name) - 1)) {
        SHA1_Update(&context, name, strlen(name));
    } else {
        RAND_bytes(name, sizeof(name) - 1);
        SHA1_Update(&context, name, sizeof(name) - 1);
    }

#if HAVE_SYS_SYSINFO_H
    if (!sysinfo(&si)) {
        SHA1_Update(&context, &si, sizeof(si));
    } else {
        RAND_bytes(name, sizeof(name) - 1);
        SHA1_Update(&context, name, sizeof(name) - 1);
    }
#else 
    /* FIXME: should find a sysinfo equiv. for this */
    RAND_bytes(name, sizeof(name) - 1);
    SHA1_Update(&context, name, sizeof(name) - 1);
#endif

    SHA1_Final(digest, &context);

    for (i = 0, ptr = NMAP.guid.next; i < SHA_DIGEST_LENGTH; i++, ptr += 2) {
        sprintf(ptr, "%02X", digest[i]);
    }

    memset(NMAP.guid.next + NMAP_GUID_PREFIX_LENGTH, '0', NMAP_GUID_SUFFIX_LENGTH);

    return;
}
#else
#error GuidReset not implemented on this platform.
#endif

void 
GuidAlloc(unsigned char *guid)
{
    int i;
    unsigned char c;
    unsigned char *ptr;

    if (guid) {
        XplWaitOnLocalSemaphore(NMAP.guid.semaphore);

        memcpy(guid, NMAP.guid.next, NMAP_GUID_LENGTH);

        ptr = NMAP.guid.next + NMAP_GUID_PREFIX_LENGTH + NMAP_GUID_SUFFIX_LENGTH - 1;
        for (i = 0; i < NMAP_GUID_SUFFIX_LENGTH; i++) {
            c = *ptr + 1;
            if (c < ':') {
                *ptr = c; /* '0' through '9' */
                break;
            } else if (c != ':') {
                if (c < 'G') {
                    *ptr = c; /* 'B' through 'F' */
                    break;
                }

                *ptr = '0';
                ptr--;
                continue;
            }

            *ptr = 'A';
            break;
        }

        if (i < NMAP_GUID_SUFFIX_LENGTH) {
            XplSignalLocalSemaphore(NMAP.guid.semaphore);

            return;
        }

        GuidReset();
        memcpy(guid, NMAP.guid.next, NMAP_GUID_LENGTH);

        XplSignalLocalSemaphore(NMAP.guid.semaphore);

        return;
    }

    return;
}

int 
NmapCommandGuid(void *param)
{
    int ccode;
    unsigned char guid[NMAP_GUID_LENGTH + 1];
    NMAPClient *client = (NMAPClient *)param;

    if (client->states & (NMAP_CLIENT_AUTHENTICATED)) {
        guid[0] = '\0';

        /* GUID */
        GuidAlloc(guid);
        guid[NMAP_GUID_LENGTH] = '\0';

        ccode = ConnWriteF(client->conn, "1000-%s\r\n", guid);
    } else {
        ccode = ConnWrite(client->conn, MSG3012BADQSTATE, sizeof(MSG3012BADQSTATE) - 1);
    }


    return(ccode);
}
