using System;
using System.Collections;
using System.Net.Sockets;
using System.Web;
using System.IO;
using System.Xml;
using Lucene.Net;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Documents;
using Lucene.Net.Index;
using Lucene.Net.Store;
using Lucene.Net.Search;
using Hula.Indexing;
using Hula.Cal;
using Hula.Store;
using Hula.Util;
using Hula.Nmap;
using Hula.Mdb;
using log4net;
using log4net.Config;

namespace Hula.Dav {

public class WrongUserException : System.Exception {
}

//<summary> 
//This is a prototype DAV service using XSP.  
//</summary>
public class DavService : IHttpHandler  {    
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    static Hula.Util.Filter nullFilter = new Hula.Util.Filter();
    
    static DavService() {
	Logging.Configure();
	log.Info("started logging");
    }

    public bool IsReusable {
	get { return true; }
    }

    public static void DecomposeUrl(string url,
				    out string owner,
				    out string store,
				    out string path)
    {
	owner = store = path = null;

	url = url.Substring(1);

	int i;
	i = url.IndexOf('/');

	if (i > -1) {
	    owner = url.Substring(0, i);
	    url = url.Substring(i + 1);
	} else {
	    owner = url;
	    store = null;
	    path = null;
	    return;
	}

	i = url.IndexOf('/');

	if (i > 0) {
	    store = url.Substring(0, i);
	    path = url.Substring(i);
	} else {
	    store = url;
	    path = null;
	}
    }    

    private string GetLocalPath(HttpRequest request, string path) {
	if (path.StartsWith (request.ApplicationPath)) {
	    return path.Substring(request.ApplicationPath.Length).TrimEnd(new char[] { '/' });
	} else {
	    return path.TrimEnd(new char[] { '/' });
	}
    }

    private Document GetDocument(LuceneIndex idx, string href) {
            TermQuery query = new TermQuery(new Term("hula.davhref", href));
            Hits hits = idx.Searcher.Search(query);
            
            if (hits == null || hits.Length() == 0) {
		return null;
            }

	    /* don't store the addition guids until we detect multiple
	     * ones, as that will be rare */
	    if (hits.Length() > 1) {
		bool fail = false;
		string guid = hits.Doc(0).GetField("guid").StringValue();
		for (int i = 1; i < hits.Length(); i++) {
		    if (guid != hits.Doc(i).GetField("guid").StringValue()) {
			fail = true;
			break;
		    }
		}

		if (fail) {
		    log.Warn (href + " returned multiple guids: ");
		    Hashtable guids = new Hashtable ();
		    for (int i = 0; i < hits.Length(); i++) {
			guid = hits.Doc(i).GetField("guid").StringValue();
			if (!guids.ContainsKey (guid)) {
			    guids[guid] = true;
			    log.Warn (" " + guid);
			}
		    }
		    throw new HttpException (500, "Internal Server Error");
		}
	    }
            
            return hits.Doc(0);
    }
 
    private void HandleOptions(HttpContext context, string user, string href) {
	string owner, storeName, path;

	DecomposeUrl(href, out owner, out storeName, out path);

	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}

	context.Response.StatusCode = 200;
	context.Response.AddHeader("Allow", 
				   "OPTIONS, GET, HEAD, PUT, DELETE");

	context.Response.AddHeader("DAV", "1");

	if (storeName == "calendar") {
	    context.Response.AddHeader("Allow", "MKCALENDAR");
	    context.Response.AddHeader("DAV", "calendar-access");
	}
    }

    private void HandleGet(HttpContext context, string user, string href) {
	string owner, storeName, path;

	DecomposeUrl(href, out owner, out storeName, out path);

	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}

	if (path == null) {
	    throw new HttpException(405, "Method not allowed");
	}

	using (LuceneIndex idx = IndexFactory.GetIndex(owner)) {
	    Document doc = GetDocument(idx, href);
	    if (doc == null) {
		throw new HttpException(404, "Not Found");
	    }

            string guid = doc.GetField("guid").StringValue();

            context.Response.StatusCode = 200;
            context.Response.AddHeader("ETag", 
                                       doc.GetField("hula.etag").StringValue());

            using (IStore store = StoreFactory.GetStore(owner, storeName)) {
		string contentType;
                using (TextReader reader = store.Get(guid,
						     context.Request.Headers["Accept"],
						     out contentType)) {
		    context.Response.ContentType = contentType;
                    nullFilter.Run(reader, context.Response.Output);
                }
            }
        }            
    }

    private void HandleHead(HttpContext context, string user, string href) {
	string owner, storeName, path;

	DecomposeUrl(href, out owner, out storeName, out path);
	
	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}	

	if (path == null) {
	    throw new HttpException(405, "Method not allowed");
	}

	LuceneIndex idx = null;
	IStore store = null;

	try {
	    idx = IndexFactory.GetIndex(owner);
	    store = StoreFactory.GetStore (owner, storeName);

	    Document doc = GetDocument(idx, href);
	    if (doc == null) {
		throw new HttpException(404, "Not Found");
	    }

            context.Response.StatusCode = 200;
            context.Response.ContentType = store.GetType (doc.GetField("hula.mime-type").StringValue(),
							  context.Request.Headers["Accept"]);	    
            context.Response.AddHeader("ETag", 
                                       doc.GetField("hula.etag").StringValue());
        } finally {
	    if (idx != null) {
		idx.Dispose();
	    }
	    if (store != null) {
		store.Dispose ();
	    }
	}
    }

    private void Index(IStore store, string guid) {
	try {
            log.Debug("Contacting Indexing server");
            NmapConnection nmap = new NmapConnection("localhost", 8689);
	    try {
                log.Debug("Pinging Indexing server");
                nmap.WriteRequest("INDEX {0} {1} Interactive {2}", 
                                  store.Owner,
                                  store.Name,
                                  guid);
                nmap.ReadReply();
	    } finally {
		nmap.Close();
	    }
	} catch (Exception e) {
	    log.Warn("Failed to connect to the indexing server: data may not be indexed");
	    throw new HttpException(500, "Failed to connect to indexing server", e);
	}
    }

    private void SplitPath(string path, out string folder, out string filename) {
        int i = path.LastIndexOf('/');
        if (i > 0) {
            folder = path.Substring(0, i);
            filename = path.Substring(i + 1);
        } else {
            folder = "/";
            filename = path.Substring(1);
        }
    }

    private void HandleConflict (HttpContext context, 
				 IStore store,
				 string guid, 
				 Document doc) {
	string contentType;

	if (store.Name != "calendar" ||
	    context.Request.Headers["Accept"] == null) {
	    goto throwException;
	}

	/* this should check that the user has permission to read the
	 * object as well*/
	try {
	    using (TextReader reader = store.Get (guid, context.Request.Headers["Accept"], out contentType)) {
		context.Response.StatusCode = 409;
		context.Response.ContentType = contentType;
		context.Response.AddHeader("ETag", 
					   doc.GetField("hula.etag").StringValue());
		
		nullFilter.Run(reader, context.Response.Output);
	    }
	} catch (HttpException e) {
	    goto throwException;
	}
	return;

    throwException:
	throw new HttpException (409, "Conflict");
    }

    private void HandlePut(HttpContext context, string user, string href) {
	string owner, storeName, path;
 
	DecomposeUrl(href, out owner, out storeName, out path);
        
	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}

	if (path == null) {
	    throw new HttpException(405, "Method not allowed");
	}

	LuceneIndex idx = null;
        IStore store = null;

        try {
            idx = IndexFactory.GetIndex(owner);
            store = StoreFactory.GetStore(owner, storeName);

            string guid;
	    string etag;
            int statusCode;

	    string match     = context.Request.Headers["If-Match"];
	    string nonematch = context.Request.Headers["If-None-Match"];

	    Document doc = GetDocument(idx, href);
            if (doc == null) {
		if (match != null && match != "*") {
		    throw new ETagException ();
		}

                // Add new document                
                string folder, filename;
                SplitPath(path, out folder, out filename);

                Hashtable properties = new Hashtable();
                properties["hula.davfilename"] = filename;
                
		store.Add(folder, 
			  properties, 
			  new StreamReader (context.Request.InputStream),
			  context.Request.ContentType,
			  out guid,
			  out etag);

                statusCode = 201;
	    } else {
		guid = doc.GetField("guid").StringValue();

		try {
		    // Replace existing document (checking ETag preconditions)
		    if (nonematch == "*") {
			throw new ETagException();
		    }
                
		    store.Replace(guid, match,
				  new StreamReader (context.Request.InputStream),
				  context.Request.ContentType,
				  out etag);
                
		    statusCode = 204;
		} catch (ETagException e) {
		    HandleConflict (context, store, guid, doc);
		    return;
		}
            }
            Index(store, guid);
	    
            context.Response.StatusCode = statusCode;
            context.Response.AddHeader("ETag", etag);
	} catch (DirectoryNotFoundException e) {
	    throw new HttpException(404, "Folder " + context.Request.Path + " was not found");
	} catch (ETagException e) {
	    /* they probably can't resolve a deleted event */
	    throw new HttpException (403, "Forbidden");
	} finally {
            if (idx != null) {
                idx.Dispose();
            }
            if (store != null) {
                store.Dispose();
            }
	}
    }

    private void HandleMkCol(HttpContext context, string user, string href) {
	string owner, storeName, path;

	DecomposeUrl(href, out owner, out storeName, out path);

	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}	

	if (path == null) {
	    throw new HttpException(405, "Method not allowed");

	}

	if (path == "/") {
	    throw new HttpException(403, "Forbidden");
	}

        string folder, filename;
        SplitPath(path, out folder, out filename);

	if (filename == null) {
	    throw new HttpException(403, "Forbidden");
	}

	using (IStore store = StoreFactory.GetStore(owner, storeName)) {
            try {
                string guid = store.AddFolder(path);
                Index(store, guid);
            } catch (DirectoryNotFoundException d) {
                throw new HttpException(409, "Conflict");
            } catch (ETagException e) {
                throw new HttpException(405, "Method not allowed");
            }
        }
    }

    private void HandleDelete (HttpContext context, string user, string href) {
	string owner, storeName, path;
 
	DecomposeUrl(href, out owner, out storeName, out path);

	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}

	if (path == null) {
	    throw new HttpException(405, "Method not allowed");
	}

	using (LuceneIndex idx = IndexFactory.GetIndex(owner)) {
	    Document doc = GetDocument(idx, href);
	    if (doc == null) {
		throw new HttpException(404, "Not Found");
	    }

	    string guid = doc.GetField("guid").StringValue();
            
	    using (IStore store = StoreFactory.GetStore(owner, storeName)) {
		try {            
		    store.Delete(guid, context.Request.Headers["If-Match"]);
		    Index(store, guid);
		    
		    context.Response.StatusCode = 204;
		} catch (ETagException e) {
		    HandleConflict (context, store, guid, doc);
		}
	    }
	}
    }

    private int GetDepth(HttpRequest r) {
        switch (r.Headers["Depth"]) {
        case "1" :
            return 1;
        case "infinity" :
            return 3;
        default :
            return 0;
        }
    }

    private void HandleReport (HttpContext context, string user, string href) {
	string owner, storeName, path;
 
	DecomposeUrl(href, out owner, out storeName, out path);

	// FIXME: we should actually do real ACL checking here, but
	// it's not implemented yet.
	if (user != owner) {
	    throw new WrongUserException ();
	}	

	int depth = GetDepth(context.Request);
	
	XmlDocument xml = new XmlDocument();
	xml.Load(context.Request.InputStream);

	using (LuceneIndex idx = IndexFactory.GetIndex(owner)) {
	    using (IStore store = StoreFactory.GetStore(owner, storeName)) {
		DavReport report = DavReport.Get(xml.DocumentElement.LocalName,
						 context);
		report.Setup(idx, store, path, depth, xml.DocumentElement);
		report.Run();
	    }
	}
    }

    public void ProcessRequest (HttpContext context) {
        string path = GetLocalPath(context.Request, context.Request.Path);
	string user = context.User.Identity.Name;

        string method = context.Request.Headers["X-Hula-HttpMethod"];
        if (method == null) {
            method = context.Request.HttpMethod;
        }

	log.DebugFormat("{0} {1}", method, context.Request.Path);
	try {
	    switch (method) {
	    case "HEAD" :
		HandleHead(context, user, path);
		break;
	    case "OPTIONS" :
		HandleOptions(context, user, path);
		break;
	    case "GET" :
		HandleGet(context, user, path);
		break;
	    case "PUT" :
		HandlePut(context, user, path);
		break;
	    case "MKCOL" :
		HandleMkCol(context, user, path);
		break;
	    case "MKCALENDAR" :
		HandleMkCol(context, user, path);
		break;
	    case "DELETE" :
		HandleDelete(context, user, path);
		break;
	    case "REPORT" :
            case "PROPFIND" :
		HandleReport(context, user, path);
		break;
	    default :
		throw new HttpException (501, "Not Implemented: " + method);
	    }
	} catch (WrongUserException e) {
	    context.Response.AppendHeader ("WWW-Authenticate", "Basic realm=\"Hula\"");
	    context.Response.StatusCode = 401;
	    context.Response.StatusDescription = "Access Denied";
	    context.Response.Write ("401 Access Denied");

	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   context.Response.StatusCode);
	    log.Warn(e);
	} catch (XmlException e) {
	    HttpException newE = new HttpException(400, "Could not parse request body", e);
	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   400);
	    log.Warn(newE);
	    throw newE;
	} catch (NmapException e) {
	    Exception newE = Hula.Nmap.NmapClient.NmapToHttpException(e);
	    log.InfoFormat("{0} {1}: {2}", 
                           method,
			   context.Request.Path,
			   400);
	    log.Warn(newE);
	    throw newE;
	} catch (HttpException e) {
	    log.InfoFormat("{0} {1}: {2}", 
                           method,
			   context.Request.Path,
			   e.GetHttpCode());
	    log.Warn(e);
	    throw e;
	} catch (NotImplementedException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   501);
	    log.Warn(e);
	    throw new HttpException(501, "Not Implemented", e);
	} catch (Exception e) {
	    log.InfoFormat("{0} {1}: {2}", 
                           method,
			   context.Request.Path,
			   500);
	    log.Warn(e);
	    throw e;
	}

	log.InfoFormat("{0} {1}: {2}", 
                       method,
		       context.Request.Path,
		       context.Response.StatusCode);
    }
}
}
