using System;
using System.Xml;
using System.Web;
using System.Text;
using System.Collections;
using System.IO;
using Lucene.Net.Search;
using Lucene.Net.Index;
using Lucene.Net.Documents;
using Hula.Store;
using Hula.Util;
using Hula.Indexing;
using Hula.Cal;
using log4net;

namespace Hula.Dav {
public abstract class QueryReport : DavReport {    
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    protected BooleanQuery query;
    
    // FIXME: We really want to check if the query matches the store.
    // For now, subclasses need to set storeMatches if the store
    // matches the query.

    protected bool storeMatches = false;

    private ArrayList filters = null;
    private ArrayList operations = null;
    private ArrayList sortFields = null;

    protected QueryReport(HttpContext context) : base(context) {
    }


    protected ArrayList Filters {
	get {
	    if (filters == null) {
		filters = new ArrayList();
	    }
	    return filters;
	}
    }
    
    protected ArrayList Operations {
	get {
	    if (operations == null) {
		operations = new ArrayList();
	    }
	    return operations;
	}
    }

    protected ArrayList SortFields {
	get {
	    if (sortFields == null) {
		sortFields = new ArrayList();
	    }
	    return sortFields;
	}
    }
    
    public BooleanQuery GetDepthQuery(IStore store, string path, int depth) {
	if (path == null || path == "") {
	    path = "/";
	}

	BooleanQuery q = new BooleanQuery();
        Query storeQuery = new TermQuery(new Term("hula.store", store.Name));
        q.Add(storeQuery, true, false);
	
        if (depth == 0) {
            Query pathQuery = new TermQuery(new Term("hula.davpath", path));
            q.Add(pathQuery, true, false);
        } else if (depth == 1) {
	    BooleanQuery subQ = new BooleanQuery();
            Query pathQuery = new TermQuery(new Term("hula.davpath", path));
            subQ.Add(pathQuery, false, false);

            Query folderQuery = new TermQuery(new Term("hula.folder", path));
            subQ.Add(folderQuery, false, false);

	    q.Add(subQ, true, false);
        } else {
	    // FIXME: do we do infinity?
	}
	return q;
    }
    
    protected virtual void SetupQuery() {}
    
    public override void Setup(LuceneIndex idx, 
			       IStore store,
			       string path,
			       int depth,
			       XmlNode node) {
	query = GetDepthQuery(store, path, depth);
	SetupQuery();
	base.Setup(idx, store, path, depth, node);
    }

    private int GetSortType(string type) {
	switch (type) {
	case "doc" :
	    return SortField.DOC;
	case "score" :
	    return SortField.SCORE;
	case "float" :
	    return SortField.STRING;
	case "int" :
	    return SortField.INT;
	case "string" :
	    return SortField.FLOAT;
	default :
	    return SortField.AUTO;
	}
    }
    
    private SortField ParseSortField (XmlNode node) {
	if (node.Attributes["field"] == null) {
	    throw new FormatException();
	}
	string name = node.Attributes["field"].Value;
	
	XmlAttribute typeAttr = node.Attributes["type"];
	int sortType = SortField.AUTO;
	if (typeAttr != null) {
	    sortType = GetSortType(typeAttr.Value);
	}

	XmlAttribute reverseAttr = node.Attributes["reverse"];
	bool reverse;
	if (reverseAttr != null) {
	    string value = reverseAttr.Value;
	    reverse = (value == "yes" || value == "1");
	} else {
	    reverse = false;
	}

	return new SortField(name,
			     sortType,
			     reverse);
    }

    protected override void HandleNode (XmlNode node) {
	switch (node.LocalName) {
	case "lucene-sort" :
	    SortFields.Add(ParseSortField(node));
	    break;
	default :
	    base.HandleNode(node);
	    break;
	}
    }

    public override void Run() {
	BeginResponse();

	// FIXME: We really want to check if the query matches the
	// store.  For now, subclasses need to set storeMatches if the
	// store matches the query.
	if (storeMatches && path == "/" || path == "" || path == null) {
	    Response(store);
	}

	Hits hits;

	log.DebugFormat("about to run {0}", query.ToString());
	
	ChainedFilter filter;
	if (filters != null && filters.Count > 0) {
	    filter = new ChainedFilter(filters, Operations);
	} else {
	    filter = null;
	}

	Sort sort;
	if (sortFields != null && sortFields.Count > 0) {
	    sort = new Sort((SortField[])sortFields.ToArray(typeof(SortField)));
	} else {
	    sort = null;
	}
	
	hits = idx.Searcher.Search(query, filter, sort);

	Responses(store, hits);

	EndResponse();
    }

    public void AddLuceneFilter(Lucene.Net.Search.Filter filter, ChainedFilter.Operation op) {
	Filters.Add(filter);
	Operations.Add(op);
    }
}
}