using System;
using System.IO;
using System.Collections;
using System.Runtime.InteropServices;

namespace Hula.ICal {

public class ICalComponent : IDisposable {
    internal const string icalLib = "ical-hula";
    public enum Kind {
	None,
	Any, /* used to select all components*/
	XRoot,
	XAttach, /* mime attached data, returned by parser. */
	VEvent,
	VTodo,
	VJournal,
	VCalendar,
	VAgenda,
	VFreebusy,
	VAlarm,
	XAudioalarm,
	XDisplayalarm,
	XEmailalarm,
	XProcedurealarm,
	VTimezone,
	XStandard,
	XSaylight,
	X,
	VXchedule,
	VQuery,
	VReply,
	VCar,
	VCommand,
	XLicInvalid,
	XLicMimepart, /* a non-stardard component that mirrors
			 structure of MIME data */
    };

    private IntPtr Handle = IntPtr.Zero;
    private string kindStr;
    bool freeHandle = true;
    
    [DllImport(icalLib)]
    private static extern IntPtr icalcomponent_new_from_string(string data);
    [DllImport(icalLib)]
    private static extern void icalcomponent_free(IntPtr component);
    [DllImport(icalLib)]
    private static extern IntPtr icalcomponent_get_first_property(IntPtr component, int propkind);
    [DllImport(icalLib)]
    private static extern IntPtr icalcomponent_get_next_property(IntPtr component, int propkind);

    [DllImport(icalLib)]
    private static extern IntPtr icalcomponent_get_first_component(IntPtr component, int propkind);
    [DllImport(icalLib)]
    private static extern IntPtr icalcomponent_get_next_component(IntPtr component, int propkind);

    [DllImport(icalLib)]
    private static extern ICalTime icalcomponent_get_dtstart(IntPtr component);
    [DllImport(icalLib)]
    private static extern ICalTime icalcomponent_get_dtend(IntPtr component);
    [DllImport(icalLib)]
    private static extern ICalDuration icalcomponent_get_duration(IntPtr component);
    [DllImport(icalLib)]
    private static extern int icalcomponent_isa(IntPtr component);
    [DllImport(icalLib)]
    private static extern IntPtr icalcomponent_kind_to_string(int kind);

    public ICalComponent(string data) {
	Handle = icalcomponent_new_from_string(data);
    }

    public ICalComponent(TextReader reader) {
	Handle = icalcomponent_new_from_string(reader.ReadToEnd());
    }

    internal ICalComponent(IntPtr handle, bool freeHandle) {
	Handle = handle;
	this.freeHandle = freeHandle;
    }    

    public ICalProperty GetFirstProperty(ICalProperty.Kind k) {
	IntPtr h = icalcomponent_get_first_property(Handle, (int)k);
	if (h != IntPtr.Zero) {
	    return new ICalProperty(h, false);
	} else {
	    return null;
	}
    }

    public ICalProperty GetNextProperty(ICalProperty.Kind k) {
	IntPtr h = icalcomponent_get_next_property(Handle, (int)k);
	if (h != IntPtr.Zero) {
	    return new ICalProperty(h, false);
	} else {
	    return null;
	}
    }
    public ICalProperty GetFirstProperty() {
	return GetFirstProperty(ICalProperty.Kind.Any);
    }

    public ICalProperty GetNextProperty() {
	return GetNextProperty(ICalProperty.Kind.Any);
    }

    public ICalComponent GetFirstComponent(Kind kind) {
	IntPtr h = icalcomponent_get_first_component(Handle, (int)kind);
	if (h != IntPtr.Zero) {
	    return new ICalComponent(h, false);
	} else {
	    return null;
	}
    }

    public ICalComponent GetNextComponent(Kind kind) {
	IntPtr h = icalcomponent_get_next_component(Handle, (int)kind);
	if (h != IntPtr.Zero) {
	    return new ICalComponent(h, false);
	} else {
	    return null;
	}
    }

    public string ComponentKind {
	get { 
	    if (kindStr == null) {
		int kind = icalcomponent_isa(Handle);
		IntPtr str = icalcomponent_kind_to_string(kind); 
		kindStr = Marshal.PtrToStringAnsi(str);
		kindStr = String.Intern(kindStr);
	    }
	    return kindStr;
	}
    }

    public IEnumerable Recurrences {
	get {
	    ICalProperty recurProp = GetFirstProperty(ICalProperty.Kind.RRule);
	    return new ICalRecurrenceEnumerable(this, recurProp);
	}
    }

    private string GetPropertyValue(ICalProperty.Kind k) {
	ICalProperty prop = GetFirstProperty(k);
	if (prop != null) {
	    return prop.Value;
	} else {
	    return null;
	}
    }

    public string Version {
	get {
	    return GetPropertyValue(ICalProperty.Kind.Version);
	}
    }

    public string ProdId {
	get {
	    return GetPropertyValue(ICalProperty.Kind.ProdId);
	}
    }

    public string Method {
	get {
	    return GetPropertyValue(ICalProperty.Kind.Method);
	}
    }

    public string CalScale {
	get {
	    return GetPropertyValue(ICalProperty.Kind.CalScale);
	}
    }

    // FIXME: really we shouldn't be touching DateTime at all, it sucks.
    public ICalTime StartTime {
	get {
	    ICalTime time = icalcomponent_get_dtstart(Handle);

	    if (time.isUtc == 0) {
		time = time.ToUniversalTime();
	    }

	    return time;
	}
    }

    public ICalTime EndTime {
	get {
	    ICalTime time = icalcomponent_get_dtend(Handle);
	    if (time.isUtc == 0) {
		time = time.ToUniversalTime();
	    }
	    return time;
	}
    }

    public ICalDuration Duration {
	get {
	    return icalcomponent_get_duration(Handle);
	}
    }

    public void Dispose() {
	if (freeHandle && Handle != IntPtr.Zero) {
	    icalcomponent_free(Handle);
	    Handle = IntPtr.Zero;
	}
	GC.SuppressFinalize(this);
    }

    ~ICalComponent() {
	Dispose();
    }
}

}
