/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/
using System;
using System.IO;
using System.Web;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Text;

using Hula.Mdb;

using log4net;

namespace Hula.Nmap
{

public class NmapClient : IDisposable
{
    private NmapConnection nconn = null;

    private AuthCommands authCmds = null;
    private MboxCommands mboxCmds = null;
    private OpenCommands openCmds = null;
    private UserCommands userCmds = null;
    private DocumentCommands docCmds = null;

    public NmapClient() : this("localhost", 689)
    {
    }

    public NmapClient(string user) : this(MsgApi.FindUserNmap(user), 689)
    {
    }

    public NmapClient(string host, int port)
    {
        nconn = new NmapConnection(host, port);

        authCmds = new AuthCommands(nconn);
        openCmds = new OpenCommands(nconn);
        userCmds = new UserCommands(nconn);
        mboxCmds = new MboxCommands(nconn);
	docCmds = new DocumentCommands(nconn);
    }

    public void Close()
    {
        if (nconn != null)
        {
            nconn.Close();
            nconn = null;
        }
    }

    public string EncodeResource(string resource) {
	return resource.Replace(' ', (char)127);
    }

    public string DecodeResource(string resource) {
	return resource.Replace((char)127, ' ');
    }

    public AuthCommands Auth
    {
        get
        {
            return authCmds;
        }
        set
        {
            this.authCmds = value;
        }
    }

    public MboxCommands Mbox
    {
        get
        {
            return mboxCmds;
        }
        set
        {
            this.mboxCmds = value;
        }
    }

    public OpenCommands Open
    {
        get
        {
            return openCmds;
        }
        set
        {
            this.openCmds = value;
        }
    }

    public UserCommands User
    {
        get
        {
            return userCmds;
        }
        set
        {
            this.userCmds = value;
        }
    }

    public DocumentCommands Document 
    {
	get
	{
	    return docCmds;
	}
	set
	{
	    this.docCmds = value;
	}
    }

    public static HttpException NmapToHttpException (NmapException e) {
	if (e.Code < 3000) {
	    // Server got a success code that it though was an error

	    return new HttpException(500, 
				     "Received a success code and interpreted it as an error", 
				     e);
	} else if (e.Code >= 3000 && e.Code < 4000) {
	    // The 3000 codes are bad-protocol codes, the server never
	    // should have caused one.
	    return new HttpException(500, 
				     "Server error while accessing NMAP", 
				     e);
	} else if (e.Code == 4000) {
	    return new HttpException (400, "Argument out of range while communicating with NMAP", e);
	} else if (e.Code == 4120 || e.Code == 4121) {
	    // FIXME: what do we throw for a lock timeout?
	    return new HttpException (500, "NMAP lock timed out", e);
	} else if (e.Code == 4220) {
	    return new HttpException (404, "Resource not found", e);
	} else if (e.Code == 4221 || e.Code == 4222 || e.Code == 4223 
		   || e.Code == 4225 || e.Code == 4227 || e.Code == 4241
		   || e.Code == 4260 || e.Code == 4261) {
	    // State errors that were implementation errors of the server
	    return new HttpException (500, "Server error while accessing NMAP", e);
	} else if (e.Code == 4224) {
	    return new HttpException(404, "Resource not found", e);
	} else if (e.Code == 4226) {
	    // ETag exceptions should have been handled earlier in the
	    // code
	    return new HttpException(500, "Server error handling conflict", e);
	} else if (e.Code == 4227 || e.Code == 4228) {
	    return new HttpException(500, "Could not write to store", e);
	} else if (e.Code == 4240) {
	    return new HttpException(401, "Permission denied", e);
	} else if (e.Code == 4930) {
	    return new HttpException(500, "NMAP server out of memory", e);
	} else {
	    return new HttpException(500, "Unknown error communicating with NMAP", e);
	}
    }


    public void Dispose() {
	Close();
    }

    ~NmapClient() {
	Close();
    }
}
}
