/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/
using System;
using System.IO;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Text;

using Hula.Mdb;

using log4net;

namespace Hula.Nmap
{
public struct NmapResponse
{
    public int code;
    public string text;
    public string data;
}

public class NmapConnection
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private TcpClient connection;
    private StreamWriter writer;
    private StreamReader reader;

    private string greeting_salt = null;

    public NmapConnection(string host, int port)
    {
        this.Open(host, port);
        this.AUTH();
    }

    public bool Open(string host, int port)
    {
        if (log.IsDebugEnabled)
        {
            log.Debug("opening NMAP connection to " + host + ":" + port);
        }

        connection = new TcpClient();
        connection.Connect(host, port);

        writer = new StreamWriter(connection.GetStream(), new ASCIIEncoding());
        reader = new StreamReader(connection.GetStream(), new ASCIIEncoding());

        NmapResponse r = this.ReadFullReply();

        if (r.code == 1000)
        {
            // no authentication will be necessary
            greeting_salt = null;
            return true;
        }
        else if (r.code == 4242)
        {
            // 4242 means we must authenticate, so save the salt
            greeting_salt = r.text;
            return true;
        }

        return false;
    }

    public void Close()
    {
        WriteRequest("QUIT");
        log.Debug("closing NMAP connection");

        connection.Close();
        connection = null;
    }

    public bool AuthAsUser (string user, string password)
    {
        this.WriteRequest ("PASS USER {0} {1}", user, password);
        if (this.ReadReply () == 1000)
        {
            return true;
        }

        return false;
    }

    public bool AUTH ()
    {
        if (greeting_salt == null)
        {
            // no auth is necessary - we're a trusted host
            return true;
        }

        string hash = MsgApi.NmapChallenge(greeting_salt);

        this.WriteRequest ("AUTH {0}", hash);
        if (this.ReadReply () == 1000)
            return true;

        this.Close ();      /* should note we've been booted? */
        return false;
    }

    public void WriteRequest (string request, params object[]args)
    {
        if (log.IsDebugEnabled)
        {
            log.Debug("SEND: " + string.Format(request, args));
            Console.Out.Flush();
        }

        writer.WriteLine (request, args);
        writer.Flush ();
    }

    public int ReadReply ()
    {
        NmapResponse response = this.ReadFullReply ();

        return response.code;
    }

    public NmapResponse ReadFullReply ()
    {
        NmapResponse response;

        string result = reader.ReadLine ();

        log.DebugFormat("RECV: {0}", result);
        Console.Out.Flush();

        response.code = Convert.ToInt32 (result.Substring (0, 4));
        response.text = result.Substring (5, result.Length - 5);
        response.data = null;

        return response;
    }

    public NmapResponse ReadMimeReply ()
    {
        NmapResponse ret = new NmapResponse();
        StringBuilder buf = new StringBuilder();

        while(true)
        {
            NmapResponse response = ReadFullReply();

            if (response.code == 1000)
            {
                ret.code = response.code;
                ret.text = response.text;
                ret.data = buf.ToString();
                return ret;
            }

            buf.Append(response.code);
            buf.Append(" ");
            buf.Append(response.text);
            buf.Append("\n");
        }
    }

    private int GetResponseLength(NmapResponse response)
    {
        string[] info = response.text.Split(' ');
	
	int chars = Convert.ToInt32(info[0]);
	if (response.code == 1000) {
	    if (info.Length > 1)
	    {
		chars = chars + Convert.ToInt32(info[1]);
	    }
	}
	
	return chars;
    }

    private NmapResponse ReadMessageData(NmapResponse response)
    {
        int chars = GetResponseLength(response);
	string data = ReadBytes(chars);
	NmapResponse ret = ReadFullReply();
	ret.data = data;

	return ret;
    }

    public string ReadBytes(int chars) {
	char[] data = new char[chars];
	int idx = 0;
	while (idx < chars) 
	{
	    int readbytes = reader.Read(data, idx, chars);
	    idx += readbytes;
	}
	return new string(data);
    } 

    public TextWriter Writer {
	get { return writer; }
    }

    public NmapStreamReader GetReplyReader ()
    {
        NmapResponse response = ReadFullReply();
	if (response.code == 2002) {
	    return new NmapStreamReader(reader, GetResponseLength (response));
	} else {
	    throw new NmapException(response);
	}
    }

    public NmapResponse ReadLongReply ()
    {
        NmapResponse response = ReadFullReply();

        if ((response.code != 2000)
            && (response.code != 2001)
            && (response.code != 2002)
            && (response.code != 2020)
            && (response.code != 2021)
            && (response.code != 2023)
            && (response.code != 2024)
            && (response.code != 2025))
        {
            return response;
        }

        StringBuilder buf = new StringBuilder();

        /* number of lines we expect to read */
        int count = -1;

        if (response.code == 2001)
        {
            buf.Append(response.text);
            buf.Append("\n");
        }
        else if ((response.code == 2002) && (!response.text.Equals("Coming up")))
        {
            count = Convert.ToInt32(response.text);
        }
        else if (response.code == 2020
                 || response.code == 2021 
                 || response.code == 2023
                 || response.code == 2024
                 || response.code == 2025)
        {
            return ReadMessageData(response);
        }

        /* number of lines read */
        int read = 0;

        while (read != count)
        {
            string line = reader.ReadLine();
            string code = line.Substring(0, 4);

            log.DebugFormat("RECV: {0}", line);
            Console.Out.Flush();

            string text = null;

            if (count == -1)
            {
                if (code.Equals("1000") || code.Equals("2002") || code.Equals("2003"))
                {
		    response.code = Int32.Parse(code);
                    response.text = line.Substring(5, line.Length - 5);
                    response.data = buf.ToString().TrimEnd();
                    return response;
                }
		else if (code.Equals("2001"))
		{
		    text = line.Substring(5, line.Length - 5);
		}
            }
            else
            {
                text = line;
            }

            buf.Append(text);
            buf.Append("\n");
            
            read++;
        }

        NmapResponse end = ReadFullReply();
        end.data = buf.ToString().TrimEnd();

        return end;
    }

    public void GetLongReplyReader (out NmapResponse response, out NmapStreamReader readerOut)
    {
        response = ReadLongReply();
	readerOut = new NmapStreamReader(reader, GetResponseLength (response));
    }

    public NmapStreamReader GetReader(int numBytes) {
	return new NmapStreamReader(reader, numBytes);
    }

    public string ReadLine() {
	return reader.ReadLine();
    }
}
}
