// Based on calconnect tests by Simon Vaillancourt

 
using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Xml;
using System.Xml.XPath;
using NUnit.Framework;
using Hula.TestData;

namespace Hula.Dav {
[TestFixture]
public class CalDavTest {
    public static int testPort = 8080;
    public static string testUser = "svaillan_ocal1_5066";
    public string testPassword = testUser;
    public static bool cleaned = false;

    public static void CleanupTestDirs() {
	// Blast the index and store out from underneath the running
	// server.
	// This is really ugly.
	DeleteDir("../../src/store-" + testUser);
	DeleteDir("../../src/index-" + testUser);
    }

    public static void DeleteDir(string dirname) {
	if (System.IO.Directory.Exists(dirname)) {
	    System.IO.Directory.Delete(dirname, true);
	}
    }

    [SetUp]
    public void SetUp() {
	if (cleaned == false) {
	    CleanupTestDirs();
	    cleaned = true;
	}
    }

    [TearDown]
    public void TearDown() {
    }

    public WebRequest GetRequest(string method, string path, string user){
	WebRequest request = WebRequest.Create("http://127.0.0.1:" + testPort.ToString() + "/dav/" + user + "/" + path);
	request.Credentials = Credentials();
	request.PreAuthenticate = true;
	request.Method = method;
	return request;
    }

    public WebRequest GetRequest(string method, string path){
	return GetRequest(method, path, testUser);
    }

    public ICredentials Credentials() {
	return new NetworkCredential(testUser, testPassword);
    }

    public void IgnoreResponse(WebRequest r) {
	r.GetResponse().Close();
    }
    public void CreateCollection(string folderName) {
	IgnoreResponse(GetRequest("MKCOL", folderName));
    }

    public void CreateCalendar(string calName) {
	string folder = "/calendar/" + calName + "/";
	CreateCollection(folder);
    }

    public WebResponse Put(string path, string ifNoneMatch, string ifMatch, string contentType, string data) {
	WebRequest request = GetRequest("PUT", path);
	if (ifNoneMatch != null) {
	    request.Headers["If-None-Match"] = ifNoneMatch;
	}
	if (ifMatch != null) {
	    request.Headers["If-Match"] = ifMatch;
	}

	request.ContentType = contentType;

	using(StreamWriter writer = new StreamWriter(request.GetRequestStream())) {
	    writer.Write(data);
	}

	return request.GetResponse();
    }

    public WebResponse Get(string path, string accept) {
	WebRequest request = GetRequest("GET", path);
	if (accept != null) {
	    ((HttpWebRequest)request).Accept = accept;
	}

	return request.GetResponse();
    }

    public WebResponse Report(string path, string data) {
	WebRequest request = GetRequest("REPORT", path);
	request.ContentLength = data.Length;
	request.ContentType = "text/xml";

	using(StreamWriter writer = new StreamWriter(request.GetRequestStream())) {
	    writer.Write(data);
	    writer.Flush();
	}

	return request.GetResponse();
    } 

    private XmlNamespaceManager CreateNamespaceManager(XmlNameTable table) {
	XmlNamespaceManager m = new XmlNamespaceManager(table);
	m.AddNamespace("D","DAV:");
	return m;
    }

    private void CheckHRefs(TextReader reader, IEnumerable hrefNames) {
	Hashtable hrefs = new Hashtable();
	foreach (string href in hrefNames) {
	    hrefs["http://127.0.0.1:8080/dav/" + testUser + href] = true;
	}

	XmlDocument doc = new XmlDocument();
	doc.Load(reader);

	XmlNamespaceManager context = CreateNamespaceManager(doc.NameTable);
	XmlNodeList nodes = doc.SelectNodes("//D:response/D:href", context);

	foreach (XmlNode node in nodes) {
	    if (hrefs.Contains(node.InnerText)) {
		hrefs.Remove(node.InnerText);
	    } else {
		Assert.Fail("unexpected href: {0}", node.InnerText);
	    }
	}
	
	if (hrefs.Count != 0) {
	    Assert.Fail("missing results");
	}
    }

    private void PutCalItem(string itemName, string filename) {
	WebResponse response = Put("/calendar/Personal/" + filename,
				   null, null, 
				   "text/calendar", Data.GetString(itemName));
	response.Close();
	   
	Assert.AreEqual(Data.GetString(itemName), GetCalItem(filename));
    }

    private string GetCalItem(string itemName) {
	WebResponse response = Get("/calendar/Personal/" + itemName, null);
	return new StreamReader(response.GetResponseStream()).ReadToEnd();
    }

    private void QueryCal(string filename, 
			  IEnumerable filenames) {
	ArrayList hrefs = new ArrayList();
	foreach (string file in filenames) {
	    hrefs.Add("/calendar/Personal/" + file);
	}
	WebResponse response = Report("/calendar/Personal/",
				      Data.GetString(filename));
	using (TextReader reader = new StreamReader(response.GetResponseStream())) {	    
	    CheckHRefs(reader, hrefs);
	}
    }
    
    // New single-instance meeting
    [Test]
    public void CreateMeeting1_1 () {
	PutCalItem("ccmeeting1_1.ics", "meeting1.ics");
    }

    // New recurring meeting
    [Test]
    public void CreateMeeting1_2 () {
	PutCalItem("ccmeeting1_2.ics", "meeting2.ics");
    }

    // New single instance meeting with 2 attendees.
    [Test]
    public void CreateMeeting1_3 () {
	PutCalItem("ccmeeting1_3.ics", "meeting3.ics");
    }

    // New single instance meeting with an alarm.
    [Test]
    public void CreateMeeting1_4 () {
	PutCalItem("ccmeeting1_4.ics", "meeting4.ics");
    }

    // Modify meeting title
    [Test]
    public void ModifyMeeting2_1 () {
	PutCalItem("ccmeeting2_1.ics", "meeting1.ics");
    }

    // Modify meeting location
    [Test]
    public void ModifyMeeting2_2 () {
	PutCalItem("ccmeeting2_2.ics", "meeting1.ics");
    }
    // Reschedule meeting to next day
    [Test]
    public void ModifyMeeting2_3 () {
	PutCalItem("ccmeeting2_3.ics", "meeting1.ics");
    }
    // Add an attendee
    [Test]
    public void ModifyMeeting2_4 () {
	PutCalItem("ccmeeting2_4.ics", "meeting1.ics");
    }
    // Add an alarm
    [Test]
    public void ModifyMeeting2_5 () {
	PutCalItem("ccmeeting2_5.ics", "meeting1.ics");
    }
    // Modify title of first instance of a recurring meeting
    [Test]
    public void ModifyMeeting2_6 () {
	PutCalItem("ccmeeting2_6.ics", "meeting2.ics");
    }
    // Modify partstat of first instance to DECLINED
    [Test]
    public void ModifyMeeting2_7 () {
	PutCalItem("ccmeeting2_7.ics", "meeting2.ics");
    }
    // Cancel 4th instance of recurring meeting
    [Test]
    public void ModifyMeeting2_8 () {
	PutCalItem("ccmeeting2_8.ics", "meeting2.ics");
    }
#if false
    // One client changes meeting time, other clients sees the modif.
    [Test]
    public void ModifyMeeting2_9 () {
	PutCalItem("ccmeeting2_9.ics", "meeting2.ics");
    }
#endif
    [Test]
    public void Query3_1_1_1() {
	QueryCal("ccquery3_1_1_1.xml",
		 new string[] {
	    "meeting1.ics",
		"meeting2.ics",
		"meeting3.ics",
		"meeting4.ics" });
    }
    [Test]
    public void Query3_1_1_2() {
	QueryCal("ccquery3_1_1_2.xml",
		 new string[] {
	    "meeting1.ics",
		"meeting2.ics",
		"meeting3.ics",
		"meeting4.ics" });
	
    }

    [Test]
    public void Query3_1_1_3() {
	QueryCal("ccquery3_1_1_3.xml",
		 new string[] {
	    "meeting1.ics",
		"meeting2.ics",
		"meeting3.ics",
		"meeting4.ics" });
	
    }

    [Test]
    public void Query3_1_2_1() {
	QueryCal("ccquery3_1_2_1.xml",
		 new string[] {
	    "meeting2.ics",
		    });
	
    }

    [Test]
    public void Query3_1_2_2() {
	QueryCal("ccquery3_1_2_2.xml",
		 new string[] {
		"meeting1.ics",
		"meeting2.ics",
		"meeting3.ics",
		"meeting4.ics",
		    });
    }
    
    [Test]
    public void Query3_1_3_1() {
#if false
	// I think this is what they were shooting for (alarms defined
	// within components) but the actual test case doesn't search for that.
	QueryCal("ccquery3_1_3_1.xml",
		 new string[] {
	         "meeting1.ics",
		"meeting4.ics",
		    });
#endif
    }

    [Test]
    public void Query3_1_3_2() {
	QueryCal("ccquery3_1_3_2.xml",
		 new string[] {
	         "meeting1.ics",
		"meeting4.ics",
		    });
	// FIXME: this doesn't actually work!
    }

    [Test]
    public void Query3_1_4_1() {
	QueryCal("ccquery3_1_4_1.xml",
		 new string[] {
		    });
    }

    [Test]
    public void Query3_1_4_2() {
	QueryCal("ccquery3_1_4_2.xml",
		 new string[] {
		    });
    }


    [Test]
    public void Query3_1_4_3() {
	QueryCal("ccquery3_1_4_3.xml",
		 new string[] {
		    });
    }

    [Test]
    public void Query3_1_5_1() {
	QueryCal("ccquery3_1_5_1.xml",
		 new string[] {
	    "meeting2.ics",
	});
    }

    [Test]
    public void Query3_1_5_2() {
	QueryCal("ccquery3_1_5_2.xml",
		 new string[] {
	    "meeting3.ics",
	});
    }

#if false

    [Test]
    public void TestReportAll() {
	SetupCalData();
	SetupAddressbookData();
	
	QueryCal("query1.xml",
		  new string[] {
	            "meeting1_6.ics",
			"meeting2_4.ics",
			"meeting3_1.ics",
			"meeting4_1.ics",
			"meeting5_1.ics" });
    }

    [Test]
    public void TestReportEvent() {
	SetupCalData();
	
	// Add a TODO 
	PutCalItem("todo1.ics");

	QueryCal("query3_1.xml",
		  new string[] {
	            "todo1.ics" });
	
	QueryCal("query3.xml",
		  new string[] {
	            "meeting1_6.ics",
			"meeting2_4.ics",
			"meeting3_1.ics",
			"meeting4_1.ics",
			"meeting5_1.ics" });
    }

    [Test]
    public void TestTimeQuery() {
	SetupCalData();
	QueryCal("query6.xml",
		     new string[] {
	            "meeting1_6.ics",
			"meeting2_4.ics",
			"meeting3_1.ics",
			"meeting4_1.ics"});

	QueryCal("query6_1.xml",
		     new string[] {
	             "meeting5_1.ics"});
    }

    [Test]
    public void TestRecurringTimeQuery() {
	SetupCalData();
	QueryCal("query5.xml",
		     new string[] { "meeting2_4.ics" });
    }

    [Test]
    public void TestPropTextMatchQuery() {
	SetupCalData();
	QueryCal("query10.xml",
		     new string[] { "meeting5_1.ics" });
    }

    [Test]
    public void TestPropTextMatchWrongCaseQuery() {
	SetupCalData();
	QueryCal("query10.1.xml",
		 new string[] { });
    }

    [Test]
    public void TestSimpleParamQuery() {
	SetupCalData();
	QueryCal("query12.xml",
		 new string[] { "meeting2_4.ics" });
    }

    [Test]
    public void TestVCardPut() {
	CreateCollection("/addressbook/TestContacts/");

	Put("/addressbook/TestContacts/card1.vcf",
	    "*", null,
	    "text/x-vcard", Data.GetString("card1.vcf")).Close();
    }

    [Test]
    public void TestVCardQuery() {
	SetupAddressbookData();
	QueryAddressbook("lucene-query1.xml",
			 new string[] { "card1.vcf" });
	QueryAddressbook("lucene-query2.xml",
			 new string[] { "card2.vcf" });
	QueryAddressbook("lucene-query3.xml",
			 new string[] { "card1.vcf", "card2.vcf" });
    }
#endif
}
}
